# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2022 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import ast
import copy
import json
import math
import os.path
import struct
import sys
import traceback

import grpc
import pyarrow
import pyarrow.flight
import numpy
from copy import deepcopy

try:
  from collections.abc import Iterable
except ImportError:
  from collections import Iterable

from datetime import datetime, date, time
from dateutil.parser import parse as parse_datetime
from . import DataService_pb2 as data_proto
from . import DataService_pb2_grpc as data_grpc
from . import GraphTypesService_pb2 as graph_proto
from . import SchemaMessages_pb2 as sch_proto
from .common import (_assert_isstring, _assert_noerrors,
                     _get_valid_property_types_for_return_only,
                     _get_valid_property_types_to_create,
                     _validated_property_name,
                     _validated_property_type,
                     _convert_flight_server_error_into_xgt,
                     _create_flight_ticket, _get_data_python, _get_data_pandas,
                     _get_data_arrow,
                     XgtNotImplemented, XgtConnectionError,
                     XgtIOError, XgtValueError, XgtInternalError, XgtTypeError,
                     MAX_PACKET_SIZE,
                     BOOLEAN, INT, FLOAT, DATE, TIME, DATETIME, IPADDRESS, TEXT, LIST)
from .job import Job

class HeaderMode:
  NONE = 'none'
  IGNORE = 'ignore'
  NORMAL = 'normal'
  STRICT = 'strict'

  _all = [NONE,IGNORE,NORMAL,STRICT]

def _validate_row_level_labels_for_ingest(row_labels,
                                          row_label_columns = None):
  if row_labels is not None and row_label_columns is not None:
    error = ("Only one of row_labels and "
             "row_label_columns must be passed."
            )
    raise ValueError(error)
  if row_labels is not None and \
     not isinstance(row_labels, (list, tuple)):
    raise TypeError('row_labels must be a list or tuple of string labels.')
  if row_label_columns is not None and \
     not isinstance(row_label_columns, (list, tuple)):
    raise TypeError('row_label_columns must be a list or tuple.')

def _validate_column_mapping_in_ingest(frame_to_file_column_mapping):
  if frame_to_file_column_mapping is not None:
    for frame_col, file_col in frame_to_file_column_mapping.items():
      if not isinstance(file_col, (str, int)):
        error_string = 'The data type of "frame_to_file_column_mapping" ' + \
                       'is incorrect. Expects a dictionary with string ' + \
                       'keys and string or integer values.'
        raise TypeError(error_string)

def _set_column_mapping_in_ingest_request(request, frame_to_file_column_mapping):
  if frame_to_file_column_mapping is not None:
    request.column_mapping.has_mapping = True
    for frame_col, file_col in frame_to_file_column_mapping.items():
      if isinstance(file_col, str):
        name_mapping = data_proto.FrameColumnToSourceName()
        name_mapping.frame_column_name = _validated_property_name(frame_col)
        name_mapping.file_column_name = file_col
        request.column_mapping.frame_column_to_source_name.extend([name_mapping])
      elif isinstance(file_col, int):
        idx_mapping = data_proto.FrameColumnToSourceIdx()
        idx_mapping.frame_column_name = _validated_property_name(frame_col)
        idx_mapping.file_column_idx = file_col
        request.column_mapping.frame_column_to_source_idx.extend([idx_mapping])
      else:
        error_string = 'The data type of "frame_to_file_column_mapping" ' + \
                       'is incorrect. Expects a dictionary with string ' + \
                       'keys and string or integer values.'
        raise TypeError(error_string)
  return request

def _get_processed_row_level_label_columns(row_label_columns,
                                           header_mode = HeaderMode.NONE):
  if row_label_columns is None:
    return None
  elif header_mode == HeaderMode.NONE or header_mode == HeaderMode.IGNORE:
    return [int(col) for col in row_label_columns]
  elif header_mode == HeaderMode.NORMAL or header_mode == HeaderMode.STRICT:
    for col in row_label_columns:
      _assert_isstring(col)
    return row_label_columns

def _row_level_labels_helper(request, row_labels, row_label_columns,
                             source_vertex_row_labels, target_vertex_row_labels,
                             header_mode):
  if row_labels is not None:
    if len(row_labels) == 0:
      row_labels.append("")
    for label in row_labels:
      request.row_labels.labels.extend([label])
  if source_vertex_row_labels is not None:
    for label in source_vertex_row_labels:
      request.row_labels.implicit_source_vertex_labels.extend([label])
  if target_vertex_row_labels is not None:
    for label in target_vertex_row_labels:
      request.row_labels.implicit_target_vertex_labels.extend([label])
  if row_label_columns is not None:
    if header_mode == HeaderMode.NONE or header_mode == HeaderMode.IGNORE:
      for col in row_label_columns:
        request.row_labels.label_column_indices.extend([col])
    elif header_mode == HeaderMode.NORMAL or header_mode == HeaderMode.STRICT:
      for col in row_label_columns:
        request.row_labels.label_column_names.extend([col])
  return request

# -----------------------------------------------------------------------------

class TableFrame(object):
  """
  TableFrame object represent a table held on the xGT server. It can be
  used to retrieve information about it and should not be instantiated directly
  by the user. Methods that return this object: `Connection.get_table_frame()`,
  `Connection.get_table_frames()` and `Connection.create_table_frame()`. A table
  may also be created by a MATCH query and may contain query results.

  Parameters
  ----------
  conn : Connection
    An open connection to an xGT server.
  name : str
    Fully qualified name of the frame, including the namespace.
  schema : list of pairs
    List of pairs associating property names with xGT data types.

  Examples
  --------
  >>> import xgt
  >>> conn = xgt.Connection()
  >>> ... run query and store results in Results
  >>> t = conn.get_table_frame('Results')
  >>> print(t.name)

  """
  def __init__(self, conn, name, schema):
    """
    Constructor for TableFrame. Called when TableFrame is created.
    """
    self._conn = conn
    self._name = name
    # Check the schema against the valid property types.
    valid_prop_types = _get_valid_property_types_to_create() + \
                       _get_valid_property_types_for_return_only()

    for col in schema:
      if col[1] not in valid_prop_types:
        raise XgtTypeError('Invalid property type "' + col[1] + '"')

    self._schema = schema

  @property
  def name(self):
    """str: Name of the frame."""
    return self._name

  @property
  def schema(self):
    """list of lists: Gets the property names and types of the frame."""
    return self._schema

  @property
  def connection(self):
    """Connection object: The connection used when constructing the frame."""
    return self._conn

  def load(self, paths,
           header_mode = HeaderMode.NONE, headerMode = None,
           record_history = True,
           row_labels = None, row_label_columns = None, delimiter = ',',
           frame_to_file_column_mapping = None, suppress_errors = False):
    """
    Loads data from one or more files specified in the list of paths.
    These files may be CSV, Parquet, or compressed CSV.
    Some limitations exist for Parquet and compressed CSV.
    See docs.trovares.com for more details.
    Each path may have its own protocol as described below.

    Parameters
    ----------
    paths : list or string
      A single path or a list of paths to files.

      ==================== =====================================
                      Syntax for one file path
      ----------------------------------------------------------
          Resource type                 Path syntax
      ==================== =====================================
          local to Python: '<path to file>'
                           'xgt://<path to file>'
          xGT server:      'xgtd://<path to file>'
          AWS s3:          's3://<path to file>'
          https site:      'https://<path to file>'
          http site:       'http://<path to file>'
          ftps server:     'ftps://<path to file>'
          ftp server:      'ftp://<path to file>'
      ==================== =====================================

    header_mode : str
      Indicates how the file header should be processed:
        - HeaderMode.NONE:
          No header exists.
        - HeaderMode.IGNORE:
          Ignore the first line containing the header.
        - HeaderMode.NORMAL:
          Process the header in non-strict mode. If a schema column is missing,
          a null value is ingested for that schema column. Any file column whose
          name does not correspond to a schema column or a security label column
          is ignored.
        - HeaderMode.STRICT:
          Process the header in strict mode. The name of each header column
          should correspond to a schema column, a security label column, or be
          named IGNORE. Each schema column must appear in the file.

      Optional. Default=HeaderMode.NONE.
      Only applies to CSV files.
      (since version 1.11.0)

    headerMode : str
      Indicates how the file header should be processed:
        - HeaderMode.NONE:
          No header exists.
        - HeaderMode.IGNORE:
          Ignore the first line containing the header.
        - HeaderMode.NORMAL:
          Process the header in non-strict mode. If a schema column is missing,
          a null value is ingested for that schema column. Any file column whose
          name does not correspond to a schema column or a security label column
          is ignored.
        - HeaderMode.STRICT:
          Process the header in strict mode. The name of each header column
          should correspond to a schema column, a security label column, or be
          named IGNORE. Each schema column must appear in the file.

      Optional. Default=HeaderMode.NONE.
      Only applies to CSV files.
      Same as header_mode. (DEPRECATED)

    record_history : bool
      If true, records the history of the job.
      (since version 1.4.0)

    row_labels : list
      A list of security labels to attach to each row inserted with the load.
      Each label must have been passed in to the row_label_universe
      parameter when creating the frame. Note: Only one of row_labels
      and row_label_columns must be passed.
      (since version 1.5.0)

    row_label_columns: list
      A list of columns indicating which columns in the CSV file contain
      security labels to attach to the inserted row. If the header mode is
      NONE or IGNORE, this must be a list of integer column indices. If the
      header mode is NORMAL or STRICT, this must be a list of string column
      names. Note: Only one of row_labels and
      row_label_columns must be passed.
      (since version 1.5.0)

    delimiter : str
      Delimiter for CSV data.
      Only applies to CSV files.
      (since version 1.5.1)

    frame_to_file_column_mapping : dictionary
      Maps the frame column names to file columns for the ingest.
      The key of each element is frame's column name. The value is either the
      name of the file column (from the file header) or the file column index.
      If file column names are used, the header_mode must be NORMAL. If only
      file column indices are used, the header_mode can be NORMAL, NONE, or
      IGNORE.
      (Beta since version 1.10)

    suppress_errors : bool
      If true, continues to load data if an ingest error is encountered,
      placing the first 1000 errors into the job history.
      If false, stops on first error and raises. Defaults to False.
      (since version 1.11.0)

    Returns
    -------
    Job
      A Job object representing the job that has executed the load.

    Raises
    -------
    XgtIOError
      If a file specified cannot be opened or if there are errors inserting any
      lines in the file into the frame.
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.

    """
    return self._load(paths, header_mode, headerMode, record_history, row_labels,
                      row_label_columns, delimiter = delimiter,
                      frame_to_file_column_mapping = frame_to_file_column_mapping,
                      suppress_errors = suppress_errors)

  def _load(self, paths, header_mode = HeaderMode.NONE, headerMode = None,
            record_history = True,
            row_labels = None, row_label_columns = None,
            source_vertex_row_labels = None, target_vertex_row_labels = None,
            delimiter = ',', frame_to_file_column_mapping = None,
            suppress_errors = False, **kwargs):
    if headerMode is not None:
      if header_mode is not HeaderMode.NONE:
        raise ValueError('Cannot pass the header mode in both the "header_mode"' +
                         ' and "headerMode" parameters. "headerMode" is deprecated.')
      else:
        header_mode = headerMode

    if not isinstance(delimiter, str) or len(delimiter) != 1:
      error = ("Delimiter must be a single character string.")
      raise ValueError(error)

    if header_mode is None:
      raise TypeError('the "header_mode" parameter is None')
    if header_mode not in HeaderMode._all:
      raise TypeError('Invalid header mode: "{0}"'.format(str(header_mode)))

    if header_mode == HeaderMode.STRICT and frame_to_file_column_mapping is not None:
      error = ("Passing frame_to_file_column_mapping with HeaderMode.STRICT "
               "is not supported.")
      raise XgtValueError(error)

    _validate_column_mapping_in_ingest(frame_to_file_column_mapping)
    _validate_row_level_labels_for_ingest(row_labels,
                                          row_label_columns)
    _validate_row_level_labels_for_ingest(source_vertex_row_labels)
    _validate_row_level_labels_for_ingest(target_vertex_row_labels)
    row_label_columns = _get_processed_row_level_label_columns(
        row_label_columns, header_mode)

    if paths is None:
      raise TypeError('the "paths" parameter is None')
    if not isinstance(paths, (list, tuple, str)):
      raise TypeError('one or more file paths are expected; the data type of the "paths" parameter is "{0}"'.format(type(paths)))
    client_paths, server_paths, url_paths = _group_paths(paths, True)
    if len(client_paths) == 0 and len(server_paths) == 0 and len(url_paths) == 0:
      raise XgtIOError('no valid paths found: ' + str(paths))
    if len(client_paths) > 0:
      return self._local_ingest(client_paths, header_mode, row_labels,
                                row_label_columns, source_vertex_row_labels,
                                target_vertex_row_labels, delimiter,
                                frame_to_file_column_mapping,
                                suppress_errors = suppress_errors)
    if len(server_paths) > 0:
      return self._ingest(server_paths, header_mode, row_labels,
                          row_label_columns, source_vertex_row_labels,
                          target_vertex_row_labels,
                          delimiter, frame_to_file_column_mapping,
                          record_history = record_history,
                          suppress_errors = suppress_errors, **kwargs)
    if len(url_paths) > 0:
      return self._ingest(url_paths, header_mode, row_labels, row_label_columns,
                          source_vertex_row_labels, target_vertex_row_labels,
                          delimiter, frame_to_file_column_mapping,
                          record_history = record_history,
                          suppress_errors = suppress_errors, **kwargs)

  def save(self, path, offset = 0, length = None, headers = False,
           record_history = True, include_row_labels = False,
           row_label_column_header = None, preserve_order = False,
           number_of_files = 1):
    """
    Writes the rows from the frame to a file in the location indicated
    by the path parameter. Will save as a Parquet file if the
    extension is .parquet, otherwise saves as a CSV.

    Parameters
    ----------
    path : str
      Path to a file.

      ==================== =====================================
                      Syntax for one file path
      ----------------------------------------------------------
          Resource type                 Path syntax
      ==================== =====================================
          local to Python: '<path to file>'
                           'xgt://<path to file>'
          xGT server:      'xgtd://<path to file>'
      ==================== =====================================

    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved.
      Optional. Default=None.
    headers : boolean
      Indicates if headers should be added.
      Optional. Default=False.
    record_history : bool
      If true, records the history of the job.
      (since version 1.4.0)
    include_row_labels : bool
      Indicates whether the security labels for each row
      should be egested along with the row.
      (since version 1.5.0)
    row_label_column_header : str
      The header column name to use for all row labels
      if include_row_labels is true and headers is true.
      (since version 1.5.0)
    preserve_order : boolean
      Indicates if the output should keep the order the
      frame is stored in.
      Optional. Default=False.
      (since version 1.5.1)
    number_of_files : int
      Number of files to save. Only works with the xgtd:// protocol.
      Optional. Default=1.
      (since version 1.10.0)

    Returns
    -------
    Job
      A Job object representing the job that has executed the save.

    Raises
    -------
    XgtIOError
      If a file specified cannot be opened.
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.

    """
    return self._save(path, offset, length, headers, record_history,
                      include_row_labels,
                      row_label_column_header, preserve_order, number_of_files)

  def _save(self, path, offset = 0, length = None, headers = False,
            record_history = True, include_row_labels = False,
            row_label_column_header = None, preserve_order = False,
            number_of_files = 1, **kwargs):
    if path is None:
      raise TypeError('the "paths" parameter is None')
    if not isinstance(path, str):
      msg = 'a file path is expected; the data type of the "path" ' \
            'parameter is "{0}"'.format(type(path))
      raise TypeError(msg)

    client_paths, server_paths, url_paths = _group_paths(path, False)
    if (len(client_paths) == 0 and len(server_paths) == 0 and
        len(url_paths) == 0):
      raise XgtIOError('no valid paths found: ' + str(path))
    if len(client_paths) > 0:
      return self._local_egest(client_paths[0], offset, length, headers,
                         include_row_labels = include_row_labels,
                         row_label_column_header = row_label_column_header,
                         preserve_order = preserve_order)
    if len(server_paths) > 0:
      return self._egest(server_paths[0], offset, length, headers,
                         record_history = record_history,
                         include_row_labels = include_row_labels,
                         row_label_column_header = row_label_column_header,
                         preserve_order = preserve_order,
                         number_of_files = number_of_files, **kwargs)

  def insert(self, data, row_labels = None,
             row_label_columns = None,
             suppress_errors = False):
    """
    Inserts data rows. The properties of the new data must match the schema
    in both order and type.

    Parameters
    ----------
    data : list or Pandas dataframe
      Data represented by a list of lists of data items or by a
      Pandas Dataframe.

    row_labels : list
      A list of security labels to attach to each row inserted.
      Each label must have been passed in to the row_label_universe
      parameter when creating the frame. Note: Only one of row_labels
      and row_label_columns must be passed.
      (since version 1.5.0)

    row_label_columns : list
      A list of integer column indices indicating which columns in the input
      data contain security labels to attach to the inserted row. Note: Only
      one of row_labels and row_label_columns must be passed.
      (since version 1.5.0)

    suppress_errors : bool
      If true, will continue to insert data if an ingest error is encountered,
      placing the first 1000 errors in the job history.  Defaults to False.
      (since version 1.11.0)

    Returns
    -------
    Job
      A Job object representing the job that has executed the insert.

    Raises
    -------
    XgtIOError
      If there are errors in the data being inserted or some data could
      not be inserted into the frame.
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.

    """
    if data is None:
      return
    if len(data) == 0:
      return

    table, invalid_rows = self._build_arrow_table(data, row_labels, row_label_columns,
                                                  suppress_errors)
    path = self._build_flight_path(row_labels, row_label_columns, suppress_errors)
    return self._write_table_to_flight(table, path, invalid_rows)

  def _build_arrow_table(self, data, row_labels,
                         row_label_columns, suppress_errors):
    def handle_error(row, invalid_rows):
      if suppress_errors:
        invalid_rows.append(row)
      else:
        raise XgtIOError(row[0])

    def validate_list_type(is_pandas, data_list, pos, i, col, data_type, schema_type, invalid_rows,
                           schema, depth = 0):
      for j, child_list in enumerate(data_list):
        if child_list is None:
          continue
        elif depth == schema[col][3] - 1:
          if not validate_value(data_list, None, is_pandas, data_list, pos, j, col,
                                type_mapping[schema[col][2]], schema[col][2],
                                invalid_rows, schema, depth + 1):
            return False
        else:
          if not validate_list_type(is_pandas, child_list, pos, j, col, data_type, schema_type,
                                    invalid_rows, schema, depth + 1):
            return False
      return True

    def validate_value(out_col, reuse, is_pandas, row, pos, i, col, data_type, schema_type,
                       invalid_rows, schema, depth = 0):
      result = row[i]
      try:
        if row[i] is None:
          pass
        elif data_type == LIST:
          result = deepcopy(result)
          if len(schema[col]) == 3 or depth == schema[col][3] - 1:
            for j, _ in enumerate(row[i]):
              if not validate_value(result, None, is_pandas, result, pos, j, col,
                                    type_mapping[schema[col][2]],
                                    schema[col][2], invalid_rows, schema,
                                    depth + 1):
                return False
          elif not validate_list_type(is_pandas, result, pos, i, col, data_type,
                                      schema_type, invalid_rows, schema, depth):
            return False
        elif not isinstance(row[i], data_type):
          if isinstance(row[i], str):
            if schema_type in { INT }:
              try:
                result = int(result, 0)
              except:
                handle_error(["Can't convert int.", "", i], invalid_rows)
                return False
            elif schema_type in { FLOAT }:
              try:
                result = float(result)
              except:
                handle_error(["Can't convert float.", "", i],
                             invalid_rows)
                return False
            elif schema_type in { DATE, DATETIME }:
              try:
                result = parse_datetime(result)
              except:
                handle_error(["Can't convert datetime.", "", i],
                             invalid_rows)
                return False
            elif schema_type in { TIME }:
              try:
                result = parse_datetime(result).time()
              except:
                handle_error(["Can't convert time.", "", i],
                             invalid_rows)
                return False
            else:
              handle_error(["Can't convert type to " + str(data_type) +
                           ".", "", i], invalid_rows)
              return False
          else:
            if not is_pandas:
              handle_error(["Can't convert type to " + str(data_type) +
                           ".", "", i], invalid_rows)
              return False
            elif row[i] is pandas.NA:
              result = None

      except Exception as err:
        handle_error([str(err), "", i], invalid_rows)
        return False
      # If reuse is none we are storing in row and not a column.
      if reuse is None:
        out_col[i] = result
      elif not reuse:
        out_col.append(result)
      else:
        out_col[-1] = result
      return True

    def create_columns(out_dict, reuse, is_pandas, row, python_types, invalid_rows, pos,
                       schema):
      if len(row) != len(python_types):
        handle_error(["Expected " + str(len(python_types)) +
                      " columns but found " + str(len(row)) + ".", "",
                      pos] + row, invalid_rows)
        return reuse
      for i, data_type in enumerate(python_types):
        if not validate_value(out_dict[i], reuse, is_pandas, row, pos, i, i, data_type,
                              schema[i][1], invalid_rows, schema):
          if not reuse and i > 0:
            # This column failed and now everything is messed up with previous
            # cols. So just append the remaining columns and we just reuse this
            # position in each column.
            for x in range(i, len(python_types)):
              out_dict[x].append(None)
            return True
          return reuse
      return False

    def create_arrow_list_type(data_type, depth):
      list_type = data_type
      for i in range(0, depth):
        list_type = pyarrow.list_(list_type)

      return list_type

    def create_arrow_type(mapping, schema, i):
      schema_length = len(schema[i])
      if schema_length == 2:
        return mapping[schema[i][1]]
      elif schema_length == 3:
        return pyarrow.list_(mapping[schema[i][2]])
      elif schema_length == 4:
        return create_arrow_list_type(mapping[schema[i][2]], schema[i][3])

    #---- if user passed in a pandas frame
    is_pandas = False
    try:
      import pandas
      if isinstance(data, pandas.DataFrame):
        is_pandas = True
    except:
      pass

    # Exceptions for iterators get eaten by grpc so we check outside
    # the generator function:
    if not is_pandas:
      if not isinstance(data, Iterable):
        raise TypeError('a list of lists or a Pandas DataFrame is expected')
      for i in range(len(data)):
          if not isinstance(data[i], Iterable):
              msg = 'Row #{0} is not a list. A list of lists ' \
                    'or a Pandas DataFrame is expected'.format(i)
              raise TypeError(msg)

    _validate_row_level_labels_for_ingest(row_labels,
                                          row_label_columns)
    row_label_columns = _get_processed_row_level_label_columns(
        row_label_columns, HeaderMode.NONE)

    # List types are complicated so for those we use a special function to
    # convert.
    arrow_mapping = { INT : pyarrow.int64(), BOOLEAN : pyarrow.bool_(),
                      TEXT : pyarrow.string(), FLOAT : pyarrow.float32(),
                      TIME : pyarrow.time64('us'), DATE : pyarrow.date32(),
                      DATETIME : pyarrow.timestamp('us'),
                      IPADDRESS : pyarrow.string(),
                      LIST : LIST }
    type_mapping = { INT : int, BOOLEAN : bool,
                     TEXT : str, FLOAT : float,
                     TIME : time, DATE : date,
                     DATETIME : datetime,
                     IPADDRESS : str,
                     LIST : LIST }

    final_schema = []

    # Locations to insert row label columns into the schema.
    if row_label_columns is not None:
      schema_mods = set(row_label_columns)
    else:
      schema_mods = { }
    correction = 0

    for i, xgt_type in enumerate(self.schema):
      # We need to insert the row label columns into the schema.
      while i + correction in schema_mods:
        correction += 1
        final_schema.append(["row_labels" + str(i + correction), TEXT] )
      final_schema.append(xgt_type)
    final_length = len(final_schema)

    # End row label columns to the end.
    for i in range(len(self.schema) + len(schema_mods) - final_length):
      final_schema.append(["row_labels" + str(i + final_length), TEXT] )

    arrow_schema_types = [create_arrow_type(arrow_mapping, final_schema, i)
                          for i, xgt_type in enumerate(final_schema)]
    python_types = [type_mapping[xgt_type[1]] for xgt_type in final_schema]
    # Table Schema (field name and type)
    arrow_schema = pyarrow.schema([pyarrow.field(final_schema[i][0], data_type)
                                   for i, data_type in
                                   enumerate(arrow_schema_types)])

    # List of invalid rows.
    invalid_rows = []
    # Columns for the final dict.
    out_dict = [[] for i in range(len(final_schema))]
    # If there is an error reuse the last appended location.
    reuse = False

    if is_pandas:
      table = data.values
    else:
      table = data

    # Here we validate, convert types, and insert into the dict columns.
    for i, row in enumerate(table):
     reuse = create_columns(out_dict, reuse, is_pandas, row, python_types, invalid_rows, i,
                            final_schema)

    # If the last row was invalid, we need to delete of the remainder.
    if reuse:
      for col in out_dict:
        col[:len(col) - 1]

    # If all the rows are garbage raise.
    if len(out_dict[0]) == 0:
      job = Job(self._conn, python_errors = invalid_rows)
      raise XgtIOError(self._create_ingest_error_message(job), job = job)

    # Create dictionary from the columns.
    arrow_dict = { final_schema[i][0] : out_dict[i] for i in
                   range(len(final_schema)) }

    return (pyarrow.Table.from_pydict(arrow_dict, arrow_schema), invalid_rows)

  def _build_flight_path(self, row_labels, row_label_columns, suppress_errors,
                         source_vertex_row_labels = None,
                         target_vertex_row_labels = None,
                         frame_to_file_column_mapping = None):
    path = tuple(self.name.split("__"))
    for x in path:
      x = '`' + x + '`'

    if row_labels is not None and len(row_labels) > 0:
      labels = ".labels=" + ",".join(f"'{w}'" for w in row_labels)
      path += (labels,)

    if row_label_columns is not None and len(row_label_columns) > 0:
      if isinstance(row_label_columns[0], int):
        label_indices = ".label_column_indices=" + \
                         ",".join(str(x) for x in row_label_columns)
        path += (label_indices,)
      elif isinstance(row_label_columns[0], str):
        label_names = ".label_column_names=" + \
                      ",".join(x for x in row_label_columns)
        path += (label_names,)

    suppress_errors_option = ".suppress_errors=" + \
                             str(suppress_errors).lower()
    path += (suppress_errors_option,)

    if source_vertex_row_labels is not None and len(source_vertex_row_labels) > 0:
      src_labels = ".implicit_source_vertex_labels=" + \
                   ",".join(f"'{w}'" for w in source_vertex_row_labels)
      path += (src_labels,)

    if target_vertex_row_labels is not None and len(target_vertex_row_labels) > 0:
      trg_labels = ".implicit_target_vertex_labels=" + \
                   ",".join(f"'{w}'" for w in target_vertex_row_labels)
      path += (trg_labels,)

    if frame_to_file_column_mapping is not None:
      _validate_column_mapping_in_ingest(frame_to_file_column_mapping)
      map_values = ".map_column_names=[" + \
          ','.join(key + ":" + value for key, value in
                   frame_to_file_column_mapping.items()
                   if isinstance(value, str)) + "]"
      path += (map_values,)
      map_values = ".map_column_ids=[" + \
          ','.join(key + ":" + str(value) for key, value in
                   frame_to_file_column_mapping.items()
                   if isinstance(value, int)) + "]"
      path += (map_values,)

    return path

  def _write_table_to_flight(self, table, path, invalid_rows):
    try:
      writer, metadata = self._conn.arrow_conn.do_put(
          pyarrow.flight.FlightDescriptor.for_path(*path), table.schema)
      writer.write_table(table)

      # Write an empty batch with metadata to indicate we are done.
      empty = [[]] * len(table.schema)
      empty_batch = pyarrow.RecordBatch.from_arrays(empty, schema = table.schema)
      metadata_end = struct.pack('<i', 0)
      writer.write_with_metadata(empty_batch, metadata_end)
      buf = metadata.read()
      job_proto = sch_proto.JobStatus()
      if buf is not None:
        job_proto.ParseFromString(buf.to_pybytes())

      writer.close()
    except pyarrow._flight.FlightServerError as err:
      raise _convert_flight_server_error_into_xgt(err) from err
    except pyarrow._flight.FlightUnavailableError as err:
      raise XgtConnectionError(str(err)) from err

    job = Job(self._conn, job_proto, python_errors = invalid_rows)
    job_data = job.get_ingest_errors()

    if job_data is not None and len(job_data) > 0:
      raise XgtIOError(self._create_ingest_error_message(job), job = job)
    return job

  def get_data(self, offset = 0, length = None, include_row_labels = False):
    """
    Returns frame data starting at a given offset and spanning a given
    length.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.
    include_row_labels : bool
      Indicates whether the security labels for each row
      should be egested along with the row.
      (since version 1.5.0)

    Returns
    -------
    list of lists

    Raises
    -------
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.
    ValueError
      If parameter is out of bounds.

    """
    ticket = _create_flight_ticket(self._name, offset, length, include_row_labels)
    return _get_data_python(self._conn, ticket)

  def get_data_arrow(self, offset = 0, length = None,
                     include_row_labels = False,
                     row_label_column_header = None):
    """
    Returns an Apache Arrow Table containing frame data starting at a given
    offset and spanning a given length.  (since version 1.11.0)

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.
    include_row_labels : bool
      Indicates whether the security labels for each row
      should be egested along with the row.
    row_label_column_header : str
      The header column name to use for all row labels
      if include_row_labels is true and headers is true.

    Returns
    -------
    Apache Arrow Table 

    Raises
    -------
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.
    ValueError
      If parameter is out of bounds.

    """
    ticket = _create_flight_ticket(self._name, offset, length, include_row_labels,
                                   row_label_column_header)
    return _get_data_arrow(self._conn, ticket)

  def get_data_pandas(self, offset = 0, length = None,
                      include_row_labels = False,
                      row_label_column_header = None):
    """
    Returns a Pandas DataFrame containing frame data starting at a given
    offset and spanning a given length.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.
    include_row_labels : bool
      Indicates whether the security labels for each row
      should be egested along with the row.
      (since version 1.5.0)
    row_label_column_header : str
      The header column name to use for all row labels
      if include_row_labels is true and headers is true.
      (since version 1.5.0)

    Returns
    -------
    Pandas DataFrame

    Raises
    -------
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.
    ValueError
      If parameter is out of bounds.

    """
    import pandas

    ticket = _create_flight_ticket(self._name, offset, length, include_row_labels,
                                   row_label_column_header)
    return _get_data_pandas(self._conn, ticket)

  @property
  def num_rows(self):
    """int: Gets the number of rows in the frame."""
    request = graph_proto.GetFrameSizeRequest()
    request.frame_name = self._name
    response = self._conn._call(request, self._conn._graph_svc.GetFrameSize)
    return response.size

  def _get_data(self, offset = 0, length = None, headers = True,
                content_type = data_proto.JSON, include_row_labels = False,
                row_label_column_header = None, preserve_order = False):
    if isinstance(offset, str):
      offset = int(offset)
    if isinstance(length, str):
      length = int(length)
    if isinstance(offset, int):
      if offset < 0:
        raise ValueError('offset is negative')
    if isinstance(length, int):
      if length < 0:
        raise ValueError('length is negative')

    request = data_proto.DownloadDataRequest()
    request.repository_name = self._name
    if offset is not None:
      request.offset.value = offset
    if length is not None:
      request.length.value = length
    request.with_headers = headers
    request.content_type = content_type
    request.preserve_order = preserve_order

    request.row_labels.egest_labels = include_row_labels
    if row_label_column_header is not None:
      request.row_labels.label_header_name = row_label_column_header
    else:
      request.row_labels.label_header_name = "ROWLABEL"

    responses = self._conn._call(request, self._conn._data_svc.DownloadData)
    return responses

  def _create_csv_packet (self, frame_name, data, header_mode, row_labels,
                          row_label_columns, source_vertex_row_labels,
                          target_vertex_row_labels, delimiter, file_path,
                          frame_to_file_column_mapping, suppress_errors):
    request = data_proto.UploadDataRequest()
    request.frame_name = frame_name.encode('utf-8')
    request.content = data.encode('utf-8')
    request.content_type = data_proto.CSV
    request.is_python_insert = False
    request.file_path = file_path
    request.suppress_errors = suppress_errors

    request = _row_level_labels_helper(request, row_labels,
                                                row_label_columns,
                                                source_vertex_row_labels,
                                                target_vertex_row_labels,
                                                header_mode)

    request.delimiter = delimiter;

    if header_mode == HeaderMode.IGNORE:
      request.header_mode = data_proto.IGNORE_HEADERS
    elif header_mode == HeaderMode.NORMAL:
      request.header_mode = data_proto.NORMAL
    elif header_mode == HeaderMode.STRICT:
      request.header_mode = data_proto.STRICT
    else:
      request.header_mode = data_proto.NONE
    request.implicit_vertices = True

    # Set the mapping of frame column to file source.
    if frame_to_file_column_mapping is not None:
      request = _set_column_mapping_in_ingest_request(request, frame_to_file_column_mapping)

    return request

  def _insert_csv_packet_generator(self, paths, header_mode, row_labels,
                                   row_label_columns, source_vertex_row_labels,
                                   target_vertex_row_labels, delimiter,
                                   frame_to_file_column_mapping,
                                   suppress_errors):
    for fpath in paths:
        try:
          header = ''
          data = ''
          dsize = 0
          with open(fpath, 'rb') as f:
            line = f.readline()
            while line:
              line = line.decode('utf-8')
              lsize = len(line)
              if (dsize + lsize) < MAX_PACKET_SIZE:
                data += line
                dsize += lsize
              else:
                yield self._create_csv_packet(self._name, data, header_mode,
                                              row_labels, row_label_columns,
                                              source_vertex_row_labels,
                                              target_vertex_row_labels,
                                              delimiter, fpath,
                                              frame_to_file_column_mapping,
                                              suppress_errors)
                data = line
                dsize = len(data)
              line = f.readline()
            yield self._create_csv_packet(self._name, data, header_mode,
                                          row_labels, row_label_columns,
                                          source_vertex_row_labels,
                                          target_vertex_row_labels, delimiter,
                                          fpath, frame_to_file_column_mapping,
                                          suppress_errors)
        except:
          #Print the error and don't throw since grpc will give an unknown error.
          sys.stderr.write("Error in " + fpath + ": ")
          traceback.print_exc(file=sys.stderr)
          sys.stderr.write("\n")
          pass

  # Returns last job run
  def _local_ingest(self, paths, header_mode=HeaderMode.NONE,
                   row_labels = None, row_label_columns = None,
                   source_vertex_row_labels = None,
                   target_vertex_row_labels = None, delimiter = ',',
                   frame_to_file_column_mapping = None,
                   suppress_errors = None):
    parquet_paths = []
    other_paths = []
    for path in paths:
      extension = os.path.splitext(path)[-1]
      if extension == ".parquet":
        parquet_paths.append(path)
      else:
        other_paths.append(path)

    # TODO (josh) : Somehow make this a single transaction for multiple paths
    #               between grpc end point and arrow flight?
    job = None
    if len(parquet_paths) > 0:
      job = self._insert_from_parquet(parquet_paths, header_mode, row_labels,
                                      row_label_columns, source_vertex_row_labels,
                                      target_vertex_row_labels,
                                      frame_to_file_column_mapping,
                                      suppress_errors)
    if len(other_paths) > 0:
      job = self._insert_from_csv(other_paths, header_mode, row_labels, row_label_columns,
                                  source_vertex_row_labels, target_vertex_row_labels,
                                  delimiter, frame_to_file_column_mapping,
                                  suppress_errors)
    return job

  # Returns last job run
  def _insert_from_parquet(self, paths, header_mode=HeaderMode.NONE,
                           row_labels = None, row_label_columns = None,
                           source_vertex_row_labels = None,
                           target_vertex_row_labels = None,
                           frame_to_file_column_mapping = None,
                           suppress_errors = None):
    flight_path = self._build_flight_path(row_labels, row_label_columns,
                                          suppress_errors,
                                          source_vertex_row_labels,
                                          target_vertex_row_labels,
                                          frame_to_file_column_mapping)
    job = None
    # TODO (josh) : Somehow make this a single transaction for
    #               multiple paths and different schemas?
    try:
      for path in paths:
        file = pyarrow.parquet.ParquetFile(path)
        writer, metadata = self._conn.arrow_conn.do_put(
          pyarrow.flight.FlightDescriptor.for_path(*flight_path),
                                                   file.schema_arrow)
        batches = file.iter_batches()
        for batch in batches:
          writer.write_batch(batch)

        # Write an empty batch with metadata to indicate we are done.
        empty = [[]] * len(file.schema_arrow)
        empty_batch = pyarrow.RecordBatch.from_arrays(
            empty, schema = file.schema_arrow)
        metadata_end = struct.pack('<i', 0)
        writer.write_with_metadata(empty_batch, metadata_end)
        buf = metadata.read()
        job_proto = sch_proto.JobStatus()
        if buf is not None:
          job_proto.ParseFromString(buf.to_pybytes())

        writer.close()
        job = Job(self._conn, job_proto)
        job_data = job.get_ingest_errors()

        if job_data is not None and len(job_data) > 0:
          raise XgtIOError(self._create_ingest_error_message(job), job = job)

    except pyarrow._flight.FlightServerError as err:
      raise _convert_flight_server_error_into_xgt(err) from err
    except pyarrow._flight.FlightUnavailableError as err:
      raise XgtConnectionError(str(err)) from err

    return job

  def _insert_from_csv(self, paths, header_mode=HeaderMode.NONE,
                       row_labels = None, row_label_columns = None,
                       source_vertex_row_labels = None,
                       target_vertex_row_labels = None, delimiter = ',',
                       frame_to_file_column_mapping = None,
                       suppress_errors = None):
    data_iter = self._insert_csv_packet_generator(paths, header_mode, row_labels,
                                                  row_label_columns,
                                                  source_vertex_row_labels,
                                                  target_vertex_row_labels,
                                                  delimiter,
                                                  frame_to_file_column_mapping,
                                                  suppress_errors)
    response = self._conn._call(data_iter, self._conn._data_svc.UploadData)
    job = Job(self._conn, response.job_status)
    job_data = job.get_ingest_errors()
    if job_data is not None and len(job_data) > 0:
      raise XgtIOError(self._create_ingest_error_message(job), job = job)
    return job

  def _ingest(self, paths, header_mode = HeaderMode.NONE,
              row_labels = None, row_label_columns = None,
              source_vertex_row_labels = None,
              target_vertex_row_labels = None, delimiter = ',',
              frame_to_file_column_mapping = None,
              suppress_errors = None, **kwargs):
    request = data_proto.IngestUriRequest()
    request.frame_name = self._name
    request = _row_level_labels_helper(request, row_labels,
                                                row_label_columns,
                                                source_vertex_row_labels,
                                                target_vertex_row_labels,
                                                header_mode)

    if isinstance(paths, (list, tuple)):
      request.content_uri.extend(paths)
    else:
      request.content_uri.extend([paths])

    if header_mode == HeaderMode.IGNORE:
      request.header_mode = data_proto.IGNORE_HEADERS
    elif header_mode == HeaderMode.NORMAL:
      request.header_mode = data_proto.NORMAL
    elif header_mode == HeaderMode.STRICT:
      request.header_mode = data_proto.STRICT
    else:
      request.header_mode = data_proto.NONE

    # Set the mapping of frame column to file source.
    if frame_to_file_column_mapping is not None:
      request = _set_column_mapping_in_ingest_request(request, frame_to_file_column_mapping)

    request.suppress_errors = suppress_errors

    if (len(self._conn.aws_access_key_id) > 0 and \
        len(self._conn.aws_secret_access_key) > 0):
      request.authorization = self._conn.aws_access_key_id + ':' + \
                              self._conn.aws_secret_access_key

    request.implicit_vertices = True
    request.delimiter = delimiter;

    for k,v in kwargs.items():
      if isinstance(v, bool):
        request.kwargs[k].bool_value = v
      elif isinstance(v, int):
        request.kwargs[k].int_value = v
      elif isinstance(v, float):
        request.kwargs[k].float_value = v
      elif isinstance(v, str):
        request.kwargs[k].string_value = v

    response = self._conn._call(request, self._conn._data_svc.IngestUri)
    job = Job(self._conn, response.job_status)
    job_data = job.get_ingest_errors()
    if job_data is not None and len(job_data) > 0:
      raise XgtIOError(self._create_ingest_error_message(job), job = job)
    return job

  def _local_egest(self, path, offset = 0, length = None, headers = False,
                   include_row_labels = False, row_label_column_header = None,
                   preserve_order = False):
    extension = os.path.splitext(path)[-1]
    if extension == ".parquet":
      return self._save_to_parquet(path, offset, length, include_row_labels,
                                   row_label_column_header, preserve_order)
    else:
      return self._save_to_csv(path, offset, length, headers, include_row_labels,
                               row_label_column_header, preserve_order)

  def _save_to_parquet(self, path, offset = 0, length = None,
                       include_row_labels = False,
                       row_label_column_header = None, preserve_order = False):
    ticket = _create_flight_ticket(self._name, offset, length, include_row_labels,
                                   row_label_column_header, preserve_order, False)

    try:
      reader = self._conn.arrow_conn.do_get(pyarrow.flight.Ticket(ticket))
      first_batch = reader.read_chunk()
      with pyarrow.parquet.ParquetWriter(path, first_batch.data.schema,
                                         version='2.6') as writer:
        writer.write_batch(first_batch.data)
        for batch in reader:
          writer.write_batch(batch.data)
    except pyarrow._flight.FlightServerError as err:
      raise _convert_flight_server_error_into_xgt(err) from err
    except pyarrow._flight.FlightUnavailableError as err:
      raise XgtConnectionError(str(err)) from err
    return Job(self._conn)

  def _save_to_csv(self, path, offset = 0, length = None, headers = False,
                   include_row_labels = False, row_label_column_header = None,
                   preserve_order = False):
    # This will stream the bytes directly which is > 10X faster than using json.
    responses = self._get_data(offset=offset, length=length, headers=headers,
                               content_type=data_proto.CSV,
                               include_row_labels = include_row_labels,
                               row_label_column_header = row_label_column_header,
                               preserve_order = preserve_order)
    job_status = None
    with open(path, 'wb') as fobject:
      # Each packet can be directly written to the file since we have the
      # raw data. This avoids extra conversion issues and extra memory from
      # json.
      for response in responses:
          _assert_noerrors(response)
          fobject.write(response.content)
          if (response.HasField("job_status")):
            if (job_status is None):
              job_status = response.job_status
            else:
              raise XgtInternalError('Job status already set in packet stream')
    fobject.close()
    return job_status

  def _egest(self, path, offset = 0, length = None, headers = False,
             include_row_labels = False, row_label_column_header = None,
             preserve_order = False, number_of_files = 1, **kwargs):
    if isinstance(offset, str):
      offset = int(offset)
    if isinstance(length, str):
      length = int(length)
    if isinstance(offset, int):
      if offset < 0:
        raise ValueError('offset is negative')
    if isinstance(length, int):
      if length < 0:
        raise ValueError('length is negative')

    request = data_proto.EgestUriRequest()
    request.frame_name = self._name
    request.file_name = path
    request.with_headers = headers
    request.preserve_order = preserve_order
    request.number_of_files = number_of_files
    request.offset.value = offset
    if length is not None:
      request.length.value = length

    request.row_labels.egest_labels = include_row_labels
    if row_label_column_header is not None:
      request.row_labels.label_header_name = row_label_column_header
    else:
      request.row_labels.label_header_name = "ROWLABEL"

    for k,v in kwargs.items():
      if isinstance(v, bool):
        request.kwargs[k].bool_value = v
      elif isinstance(v, int):
        request.kwargs[k].int_value = v
      elif isinstance(v, float):
        request.kwargs[k].float_value = v
      elif isinstance(v, str):
        request.kwargs[k].string_value = v

    response = self._conn._call(request, self._conn._data_svc.EgestUri)
    return Job(self._conn, response.job_status)

  def __str__(self):
    print_frame = "{'name': '" + self.name + "'" + \
                  ", 'schema': " + str(self.schema) + "}"
    return print_frame

  @property
  def row_label_universe(self):
    """list of strings: Gets the universe of row security labels that can be
       attached to rows of this frame. Only labels that are also in the
       authenticated user's label set are returned."""
    request = graph_proto.GetRowLabelUniverseRequest()
    request.frame_name = self.name
    response = self._conn._call(request, self._conn._graph_svc.GetRowLabelUniverse)
    return response.row_labels.label

  def _create_ingest_error_message(self, job):
    num_errors = job.total_ingest_errors
    num_rows = self.num_rows

    error_string = 'Errors occurred when inserting data into frame ' + \
                   self._name + '.\n'

    error_string += '  ' + str(num_errors) + ' line'
    if num_errors > 1:
      error_string += 's'

    error_string += ' had insertion errors.\n' + \
                    '  Lines without errors were inserted into the frame.\n' + \
                    '  To see the number of rows in the frame, run "' + \
                    self._name + '.num_rows".\n' + \
                    '  To see the data in the frame, run "' + \
                    self._name + '.get_data()".\n'

    extra_text = ''
    if num_errors > 10:
      extra_text = ' first 10'

    error_string += 'Errors associated with the' + extra_text + ' lines ' + \
                    'that could not be inserted are shown below:'

    # Only print the first 10 messages.
    for error in job.get_ingest_errors(0, 10):
      delim = ','

      # Will process a list of strings. Convert to this format.
      if isinstance(error, str):
        error_cols = error.split(delim)
      elif isinstance(error, list):
        error_cols = [str(elem) for elem in error]
      else:
        raise XgtIOError("Error processing ingest error message.")

      # The first comma separated fields of the error string are the
      # error description, file name, and line number.
      error_explanation = "" if len(error_cols) < 1 else error_cols[0]
      error_file_name = "" if len(error_cols) < 2 else os.path.basename(error_cols[1])
      error_line_number = "" if len(error_cols) < 3 else error_cols[2]

      # The second part of the error string contains the line that
      # caused the error. The line contains comma separated fields
      # so we need to re-join these comma separated portions to get back
      # the line.
      line_with_error = "" if len(error_cols) < 4 else delim.join(error_cols[3:])

      if error_line_number == -1:
        error_string += '\n '+ str(error_explanation)
      else:
        error_string += '\n  File: ' + error_file_name + ': Line: ' + str(error_line_number) + \
                        ': ' + str(error_explanation)

    return error_string

# -----------------------------------------------------------------------------

class VertexFrame(TableFrame):
  """
  VertexFrame object represents a collection of vertices held on the xGT
  server; it can be used to retrieve information about them and should not be
  instantiated directly by the user. Methods that return this object:
  `Connection.get_vertex_frame()`, `Connection.get_vertex_frames()` and
  `Connection.create_vertex_frame()`.

  Each vertex in a VertexFrame shares the same properties,
  described in `VertexFrame.schema`. Each vertex in a VertexFrame
  is uniquely identified by the property listed in `VertexFrame.key`.

  Parameters
  ----------
  conn : Connection
    An open connection to an xGT server.
  name : str
    Fully qualified name of the vertex frame, including the namespace.
  schema : list of pairs
    List of pairs associating property names with xGT data types.
    Each vertex in the VertexFrame will have these properties.
  key : str
    The property name used to uniquely identify vertices in the graph.
    This is the name of one of the properties from the schema and
    must be unique for each vertex in the frame.

  Examples
  --------
  >>> import xgt
  >>> conn = xgt.Connection()
  >>> v1 = conn.create_vertex_frame(
  ...        name = 'People',
  ...        schema = [['id', xgt.INT],
  ...                  ['name', xgt.TEXT]],
  ...        key = 'id')
  >>> v2 = conn.get_vertex_frame('Companies') # An existing vertex frame
  >>> print(v1.name, v2.name)

  """
  def __init__(self, conn, name, schema, key):
    """
    Constructor for VertexFrame. Called when VertexFrame is created.
    """
    super(VertexFrame, self).__init__(conn, name, schema)
    self._key = key

  @property
  def key(self):
    """str: Gets the property name that uniquely identifies vertices of this type."""
    return self._key

  @property
  def num_vertices(self):
    """int: Gets the number of vertices in the VertexFrame."""
    return self.num_rows

  def __str__(self):
    print_frame = ("{'name': '" + self.name + "'" +
                   ", 'schema': " + str(self.schema) +
                   ", 'key': '" + self.key + "'}")
    return print_frame

# -----------------------------------------------------------------------------

class EdgeFrame(TableFrame):
  """
  EdgeFrame object represents a collection of edges held on the xGT server;
  it can be used to retrieve information about them and should not be
  instantiated directly by the user. Methods that return this object:
  `Connection.get_edge_frame()`, `Connection.get_edge_frames()` and
  `Connection.create_edge_frame()`. Each edge in an EdgeFrame shares the same
  properties, described in `EdgeFrame.schema`.

  The source vertex of each edge in an EdgeFrame must belong to the same
  VertexFrame. This name of this VertexFrame is given by `EdgeFrame.source_name`.
  The targe vertex of each edge in an EdgeFrame must belong to the same
  VertexFrame. This name of this VertexFrame is given by `EdgeFrame.target_name`.

  For each edge in the EdgeFrame, its source vertex is identified by
  the edge property name given by `EdgeFrame.source_key`, which is
  one of the properties listed in the schema. The edge target vertex
  is identified by the property name given by `EdgeFrame.target_key`.

  Parameters
  ----------
  conn : Connection
    An open connection to an xGT server.
  name : str
    Fully qualified name of the edge frame, including the namespace.
  schema : list of pairs
    List of pairs associating property names with xGT data types.
    Each edge in the EdgeFrame will have these properties.
  source : str or VertexFrame
    The name of a VertexFrame or a VertexFrame object.
    The source vertex of each edge in this EdgeFrame will belong
    to this VertexFrame.
  target : str or VertexFrame
    The name of a VertexFrame or a VertexFrame object.
    The target vertex of each edge in this EdgeFrame will belong
    to this VertexFrame.
  source_key : str
    The edge property name that identifies the source vertex of an edge.
    This is one of the properties from the schema.
  target_key : str
    The edge property name that identifies the target vertex of an edge.
    This is one of the properties from the schema.

  Examples
  --------
  >>> import xgt
  >>> conn = xgt.Connection()
  >>> e1 = conn.create_edge_frame(
  ...        name = 'WorksFor',
  ...        schema = [['srcid', xgt.INT],
  ...                  ['role', xgt.TEXT],
  ...                  ['trgid', xgt.INT]],
  ...        source = 'People',
  ...        target = 'Companies',
  ...        source_key = 'srcid',
  ...        target_key = 'trgid')
  >>> e2 = conn.get_edge_frame('RelatedTo') # An existing edge frame
  >>> print(e1.name, e2.name)

  """
  def __init__(self, conn, name, schema, source, target, source_key, target_key):
    """
    Constructor for EdgeFrame. Called when EdgeFrame is created.
    """
    super(EdgeFrame, self).__init__(conn, name, schema)
    self._source_name = source
    self._target_name = target
    self._source_key = source_key
    self._target_key = target_key

  @property
  def source_name(self):
    """str: Gets the name of the source vertex frame."""
    return self._source_name

  @property
  def target_name(self):
    """str: Gets the name of the target vertex frame."""
    return self._target_name

  @property
  def source_key(self):
    """str: The edge property name that identifies the source vertex of an edge."""
    return self._source_key

  @property
  def target_key(self):
    """str: The edge property name that identifies the target vertex of an edge."""
    return self._target_key

  @property
  def num_edges(self):
    """int: Gets the number of edges in the EdgeFrame."""
    return self.num_rows

  def __str__(self):
    print_frame = ("{'name': '" + self.name + "'" +
                   ", 'source': '" + self.source_name + "'" +
                   ", 'target': '" + self.target_name + "'" +
                   ", 'schema': " + str(self.schema) +
                   ", 'source_key': '" + self.source_key + "'" +
                   ", 'target_key': '" + self.target_key + "'}")
    return print_frame

  def load(self, paths,
           header_mode = HeaderMode.NONE, headerMode = None,
           record_history = True,
           row_labels = None, row_label_columns = None,
           source_vertex_row_labels = None, target_vertex_row_labels = None,
           delimiter = ',', frame_to_file_column_mapping = None,
           suppress_errors = False):
    """
    Loads data from one or more files specified in the list of paths.
    These files may be CSV, Parquet, or compressed CSV.
    Some limitations exist for Parquet and compressed CSV.
    See docs.trovares.com for more details.
    Each path may have its own protocol as described below.

    Parameters
    ----------
    paths : list or string
      A single path or a list of paths to files.

      ==================== =====================================
                      Syntax for one file path
      ----------------------------------------------------------
          Resource type                 Path syntax
      ==================== =====================================
          local to Python: '<path to file>'
                           'xgt://<path to file>'
          xGT server:      'xgtd://<path to file>'
          AWS s3:          's3://<path to file>'
          https site:      'https://<path to file>'
          http site:       'http://<path to file>'
          ftps server:     'ftps://<path to file>'
          ftp server:      'ftp://<path to file>'
      ==================== =====================================

    header_mode : str
      Indicates how the file header should be processed:
        - HeaderMode.NONE:
          No header exists.
        - HeaderMode.IGNORE:
          Ignore the first line containing the header.
        - HeaderMode.NORMAL:
          Process the header in non-strict mode. If a schema column is missing,
          a null value is ingested for that schema column. Any file column whose
          name does not correspond to a schema column or a security label column
          is ignored.
        - HeaderMode.STRICT:
          Process the header in strict mode. The name of each header column
          should correspond to a schema column, a security label column, or be
          named IGNORE. Each schema column must appear in the file.

      Optional. Default=HeaderMode.NONE.
      Only applies to CSV files.
      (since version 1.11.0)

    headerMode : str
      Indicates if the CSV files contain headers:
        - HeaderMode.NONE
        - HeaderMode.IGNORE
        - HeaderMode.NORMAL
        - HeaderMode.STRICT

      Optional. Default=HeaderMode.NONE.
      Only applies to CSV files.
      Same as header_mode. (DEPRECATED)

    record_history : bool
      If true, records the history of the job.
      (since version 1.4.0)

    row_labels : list
      A list of security labels to attach to each row inserted with the load.
      Each label must have been passed in to the row_label_universe
      parameter when creating the frame. Note: Only one of row_labels
      and row_label_columns must be passed.
      (since version 1.5.0)

    row_label_columns: list
      A list of columns indicating which columns in the CSV file contain
      security labels to attach to the inserted row. If the header mode is
      NONE or IGNORE, this must be a list of integer column indices. If the
      header mode is NORMAL or STRICT, this must be a list of string column
      names. Note: Only one of row_labels and
      row_label_columns must be passed.
      (since version 1.5.0)

    source_vertex_row_labels : list
      A list of security labels to attach to each source vertex that is
      implicitly inserted. Each label must have been passed in to the
      row_label_universe parameter when creating the frame.
      (since version 1.5.0)

    target_vertex_row_labels : list
      A list of security labels to attach to each target vertex that is
      implicitly inserted. Each label must have been passed in to the
      row_label_universe parameter when creating the frame.
      (since version 1.5.0)

    delimiter : str
      Delimiter for CSV data.
      Only applies to CSV files.
      (since version 1.5.1)

    frame_to_file_column_mapping : dictionary
      Maps the frame column names to file columns for the ingest.
      The key of each element is frame's column name. The value is either the
      name of the file column (from the file header) or the file column index.
      If file column names are used, the header_mode must be NORMAL. If only
      file column indices are used, the header_mode can be NORMAL, NONE, or
      IGNORE.
      (Beta since version 1.10)

    suppress_errors : bool
      If true, continues to load data if an ingest error is encountered,
      placing the first 1000 errors into the job history.
      If false, stops on first error and raises. Defaults to False.
      (since version 1.11.0)

    Returns
    -------
    Job
      A Job object representing the job that has executed the load.

    Raises
    -------
    XgtIOError
      If a file specified cannot be opened or if there are errors inserting any
      lines in the file into the frame.
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.

    """
    return self._load(paths, header_mode, headerMode, record_history, row_labels,
                      row_label_columns, source_vertex_row_labels,
                      target_vertex_row_labels, delimiter,
                      frame_to_file_column_mapping = frame_to_file_column_mapping,
                      suppress_errors = suppress_errors)

  def insert(self, data, row_labels = None,
             row_label_columns = None,
             source_vertex_row_labels = None,
             target_vertex_row_labels = None,
             suppress_errors = False):
    """
    Inserts data rows. The properties of the new data must match the schema
    in both order and type.

    Parameters
    ----------
    data : list or Pandas dataframe
      Data represented by a list of lists of data items or by a
      Pandas Dataframe.

    row_labels : list
      A list of security labels to attach to each row inserted.
      Each label must have been passed in to the row_label_universe
      parameter when creating the frame. Note: Only one of row_labels
      and row_label_columns must be passed.
      (since version 1.5.0)

    row_label_columns: list
      A list of integer column indices indicating which columns in the input
      data contain security labels to attach to the inserted row. Note: Only
      one of row_labels and row_label_columns must be passed.
      (since version 1.5.0)

    source_vertex_row_labels : list
      A list of security labels to attach to each source vertex that is
      implicitly inserted. Each label must have been passed in to the
      row_label_universe parameter when creating the frame.
      (since version 1.5.0)

    target_vertex_row_labels : list
      A list of security labels to attach to each target vertex that is
      implicitly inserted. Each label must have been passed in to the
      row_label_universe parameter when creating the frame.
      (since version 1.5.0)

    suppress_errors : bool
      If true, continues to load data if an ingest error is encountered,
      placing the first 1000 errors into the job history.
      If false, stops on first error and raises. Defaults to False.
      (since version 1.11.0)

    Returns
    -------
    Job
      A Job object representing the job that has executed the insert.

    Raises
    -------
    XgtIOError
      If there are errors in the data being inserted or some data could
      not be inserted into the frame.
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.

    """
    if data is None:
      return
    if len(data) == 0:
      return

    _validate_row_level_labels_for_ingest(source_vertex_row_labels)
    _validate_row_level_labels_for_ingest(target_vertex_row_labels)

    table, invalid_rows = self._build_arrow_table(data, row_labels, row_label_columns,
                                                  suppress_errors)
    path = self._build_flight_path(row_labels, row_label_columns, suppress_errors,
                                   source_vertex_row_labels, target_vertex_row_labels)
    return self._write_table_to_flight(table, path, invalid_rows)

# -----------------------------------------------------------------------------

def _group_paths(paths, for_ingest):
  client_prefix = 'xgt://'
  server_prefix = 'xgtd://'
  url_prefixes = ['s3://', 'https://', 'http://', 'ftps://', 'ftp://']
  client_paths = []
  server_paths = []
  url_paths = []
  if isinstance(paths, str):
    paths = [paths]
  elif not isinstance(paths, (list, tuple)):
    return client_paths, server_paths, url_paths
  for one_path in paths:
    if one_path == "":
      raise ValueError('one or more "paths" are empty')
    if one_path.startswith(client_prefix):
      _validate_client_path(one_path)
      client_paths.append(one_path[len(client_prefix):])
    elif one_path.startswith(server_prefix):
      server_paths.append(one_path[len(server_prefix):])
    elif any(map(lambda p: one_path.startswith(p), url_prefixes)):
      for url_prefix in url_prefixes:
        if for_ingest == False:
          msg = 'Url paths are invalid for data writing ' \
                '"{0}".'.format(one_path)
          raise XgtNotImplemented(msg)
        if one_path.startswith(url_prefix):
          url_paths.append(one_path)
          break
    else:
      if '://' in one_path:
        msg = 'Unsupported url protocol in path "{0}".'.format(one_path)
        raise XgtNotImplemented(msg)
      _validate_client_path(one_path)
      client_paths.append(one_path)
  return client_paths, server_paths, url_paths

def _validate_client_path(one_path):
  if one_path.endswith('.gz') or one_path.endswith('.bz2'):
    msg = 'Loading/Saving compressed files from a local filesystem is ' \
          'not supported: {0}'.format(one_path)
    raise XgtNotImplemented(msg)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: GraphTypesService.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import AdminService_pb2 as AdminService__pb2
from . import ErrorMessages_pb2 as ErrorMessages__pb2
from . import SchemaMessages_pb2 as SchemaMessages__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17GraphTypesService.proto\x12\x03xgt\x1a\x12\x41\x64minService.proto\x1a\x13\x45rrorMessages.proto\x1a\x14SchemaMessages.proto\"\x19\n\x08LabelSet\x12\r\n\x05label\x18\x01 \x03(\t\"\x90\x01\n\rFrameLabelMap\x12;\n\raccess_labels\x18\x01 \x03(\x0b\x32$.xgt.FrameLabelMap.AccessLabelsEntry\x1a\x42\n\x11\x41\x63\x63\x65ssLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.xgt.LabelSet:\x02\x38\x01\"q\n\x17GetVertexFramesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12&\n\x0cvertex_frame\x18\x02 \x03(\x0b\x32\x10.xgt.VertexFrame\x12\x13\n\x0b\x65rror_frame\x18\x03 \x03(\t\"k\n\x15GetEdgeFramesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12\"\n\nedge_frame\x18\x02 \x03(\x0b\x32\x0e.xgt.EdgeFrame\x12\x13\n\x0b\x65rror_frame\x18\x03 \x03(\t\"n\n\x16GetTableFramesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12$\n\x0btable_frame\x18\x02 \x03(\x0b\x32\x0f.xgt.TableFrame\x12\x13\n\x0b\x65rror_frame\x18\x03 \x03(\t\"\xfe\x01\n\x16\x43reateNamespaceRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\x12+\n\x0f\x66rame_label_map\x18\x02 \x01(\x0b\x32\x12.xgt.FrameLabelMap\x12\x37\n\x06kwargs\x18\x03 \x03(\x0b\x32\'.xgt.CreateNamespaceRequest.KwargsEntry\x12)\n\x12row_label_universe\x18\x04 \x01(\x0b\x32\r.xgt.LabelSet\x1a?\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.xgt.ConfigValue:\x02\x38\x01\"4\n\x17\x43reateNamespaceResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\"\xba\x01\n\x16\x44\x65leteNamespaceRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\x12\x12\n\nforce_drop\x18\x02 \x01(\x08\x12\x37\n\x06kwargs\x18\x03 \x03(\x0b\x32\'.xgt.DeleteNamespaceRequest.KwargsEntry\x1a?\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.xgt.ConfigValue:\x02\x38\x01\"O\n\x17\x44\x65leteNamespaceResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12\x19\n\x11\x66ound_and_deleted\x18\x02 \x01(\x08\"\x16\n\x14GetNamespacesRequest\"J\n\x15GetNamespacesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12\x16\n\x0enamespace_name\x18\x02 \x03(\t\"\x8b\x03\n\x12\x43reateFrameRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\x12\x1b\n\x06schema\x18\x02 \x01(\x0b\x32\x0b.xgt.Schema\x12\x15\n\rsource_vertex\x18\x03 \x01(\t\x12\x15\n\rtarget_vertex\x18\x04 \x01(\t\x12\x12\n\nsource_key\x18\x05 \x01(\t\x12\x12\n\ntarget_key\x18\x06 \x01(\t\x12 \n\x04type\x18\x07 \x01(\x0e\x32\x12.xgt.FrameTypeEnum\x12+\n\x0f\x66rame_label_map\x18\x08 \x01(\x0b\x32\x12.xgt.FrameLabelMap\x12\x33\n\x06kwargs\x18\t \x03(\x0b\x32#.xgt.CreateFrameRequest.KwargsEntry\x12)\n\x12row_label_universe\x18\n \x01(\x0b\x32\r.xgt.LabelSet\x1a?\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.xgt.ConfigValue:\x02\x38\x01\"@\n\x18GetFramesMetadataRequest\x12\x0c\n\x04name\x18\x01 \x03(\t\x12\x16\n\x0enamespace_name\x18\x02 \x01(\t\"X\n\x19GetFramesMetadataResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12 \n\tcontainer\x18\x02 \x03(\x0b\x32\r.xgt.AnyFrame\"\x9e\x01\n\x12\x44\x65leteFrameRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\x12\x33\n\x06kwargs\x18\x02 \x03(\x0b\x32#.xgt.DeleteFrameRequest.KwargsEntry\x1a?\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.xgt.ConfigValue:\x02\x38\x01\"K\n\x13\x44\x65leteFrameResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12\x19\n\x11\x66ound_and_deleted\x18\x02 \x01(\x08\")\n\x13GetFrameSizeRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\"?\n\x14GetFrameSizeResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12\x0c\n\x04size\x18\x02 \x01(\x04\"+\n\x15GetFrameLabelsRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\"`\n\x16GetFrameLabelsResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12+\n\x0f\x66rame_label_map\x18\x02 \x01(\x0b\x32\x12.xgt.FrameLabelMap\"0\n\x1aGetRowLabelUniverseRequest\x12\x12\n\nframe_name\x18\x01 \x01(\t\"[\n\x1bGetRowLabelUniverseResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12!\n\nrow_labels\x18\x02 \x01(\x0b\x32\r.xgt.LabelSet2\xf4\x06\n\x11GraphTypesService\x12R\n\x0fGetVertexFrames\x12\x1d.xgt.GetFramesMetadataRequest\x1a\x1e.xgt.GetFramesMetadataResponse\"\x00\x12P\n\rGetEdgeFrames\x12\x1d.xgt.GetFramesMetadataRequest\x1a\x1e.xgt.GetFramesMetadataResponse\"\x00\x12Q\n\x0eGetTableFrames\x12\x1d.xgt.GetFramesMetadataRequest\x1a\x1e.xgt.GetFramesMetadataResponse\"\x00\x12N\n\x0f\x43reateNamespace\x12\x1b.xgt.CreateNamespaceRequest\x1a\x1c.xgt.CreateNamespaceResponse\"\x00\x12N\n\x0f\x44\x65leteNamespace\x12\x1b.xgt.DeleteNamespaceRequest\x1a\x1c.xgt.DeleteNamespaceResponse\"\x00\x12H\n\rGetNamespaces\x12\x19.xgt.GetNamespacesRequest\x1a\x1a.xgt.GetNamespacesResponse\"\x00\x12H\n\x0b\x43reateFrame\x12\x17.xgt.CreateFrameRequest\x1a\x1e.xgt.GetFramesMetadataResponse\"\x00\x12\x42\n\x0b\x44\x65leteFrame\x12\x17.xgt.DeleteFrameRequest\x1a\x18.xgt.DeleteFrameResponse\"\x00\x12\x45\n\x0cGetFrameSize\x12\x18.xgt.GetFrameSizeRequest\x1a\x19.xgt.GetFrameSizeResponse\"\x00\x12K\n\x0eGetFrameLabels\x12\x1a.xgt.GetFrameLabelsRequest\x1a\x1b.xgt.GetFrameLabelsResponse\"\x00\x12Z\n\x13GetRowLabelUniverse\x12\x1f.xgt.GetRowLabelUniverseRequest\x1a .xgt.GetRowLabelUniverseResponse\"\x00\x62\x06proto3')



_LABELSET = DESCRIPTOR.message_types_by_name['LabelSet']
_FRAMELABELMAP = DESCRIPTOR.message_types_by_name['FrameLabelMap']
_FRAMELABELMAP_ACCESSLABELSENTRY = _FRAMELABELMAP.nested_types_by_name['AccessLabelsEntry']
_GETVERTEXFRAMESRESPONSE = DESCRIPTOR.message_types_by_name['GetVertexFramesResponse']
_GETEDGEFRAMESRESPONSE = DESCRIPTOR.message_types_by_name['GetEdgeFramesResponse']
_GETTABLEFRAMESRESPONSE = DESCRIPTOR.message_types_by_name['GetTableFramesResponse']
_CREATENAMESPACEREQUEST = DESCRIPTOR.message_types_by_name['CreateNamespaceRequest']
_CREATENAMESPACEREQUEST_KWARGSENTRY = _CREATENAMESPACEREQUEST.nested_types_by_name['KwargsEntry']
_CREATENAMESPACERESPONSE = DESCRIPTOR.message_types_by_name['CreateNamespaceResponse']
_DELETENAMESPACEREQUEST = DESCRIPTOR.message_types_by_name['DeleteNamespaceRequest']
_DELETENAMESPACEREQUEST_KWARGSENTRY = _DELETENAMESPACEREQUEST.nested_types_by_name['KwargsEntry']
_DELETENAMESPACERESPONSE = DESCRIPTOR.message_types_by_name['DeleteNamespaceResponse']
_GETNAMESPACESREQUEST = DESCRIPTOR.message_types_by_name['GetNamespacesRequest']
_GETNAMESPACESRESPONSE = DESCRIPTOR.message_types_by_name['GetNamespacesResponse']
_CREATEFRAMEREQUEST = DESCRIPTOR.message_types_by_name['CreateFrameRequest']
_CREATEFRAMEREQUEST_KWARGSENTRY = _CREATEFRAMEREQUEST.nested_types_by_name['KwargsEntry']
_GETFRAMESMETADATAREQUEST = DESCRIPTOR.message_types_by_name['GetFramesMetadataRequest']
_GETFRAMESMETADATARESPONSE = DESCRIPTOR.message_types_by_name['GetFramesMetadataResponse']
_DELETEFRAMEREQUEST = DESCRIPTOR.message_types_by_name['DeleteFrameRequest']
_DELETEFRAMEREQUEST_KWARGSENTRY = _DELETEFRAMEREQUEST.nested_types_by_name['KwargsEntry']
_DELETEFRAMERESPONSE = DESCRIPTOR.message_types_by_name['DeleteFrameResponse']
_GETFRAMESIZEREQUEST = DESCRIPTOR.message_types_by_name['GetFrameSizeRequest']
_GETFRAMESIZERESPONSE = DESCRIPTOR.message_types_by_name['GetFrameSizeResponse']
_GETFRAMELABELSREQUEST = DESCRIPTOR.message_types_by_name['GetFrameLabelsRequest']
_GETFRAMELABELSRESPONSE = DESCRIPTOR.message_types_by_name['GetFrameLabelsResponse']
_GETROWLABELUNIVERSEREQUEST = DESCRIPTOR.message_types_by_name['GetRowLabelUniverseRequest']
_GETROWLABELUNIVERSERESPONSE = DESCRIPTOR.message_types_by_name['GetRowLabelUniverseResponse']
LabelSet = _reflection.GeneratedProtocolMessageType('LabelSet', (_message.Message,), {
  'DESCRIPTOR' : _LABELSET,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.LabelSet)
  })
_sym_db.RegisterMessage(LabelSet)

FrameLabelMap = _reflection.GeneratedProtocolMessageType('FrameLabelMap', (_message.Message,), {

  'AccessLabelsEntry' : _reflection.GeneratedProtocolMessageType('AccessLabelsEntry', (_message.Message,), {
    'DESCRIPTOR' : _FRAMELABELMAP_ACCESSLABELSENTRY,
    '__module__' : 'GraphTypesService_pb2'
    # @@protoc_insertion_point(class_scope:xgt.FrameLabelMap.AccessLabelsEntry)
    })
  ,
  'DESCRIPTOR' : _FRAMELABELMAP,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.FrameLabelMap)
  })
_sym_db.RegisterMessage(FrameLabelMap)
_sym_db.RegisterMessage(FrameLabelMap.AccessLabelsEntry)

GetVertexFramesResponse = _reflection.GeneratedProtocolMessageType('GetVertexFramesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETVERTEXFRAMESRESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetVertexFramesResponse)
  })
_sym_db.RegisterMessage(GetVertexFramesResponse)

GetEdgeFramesResponse = _reflection.GeneratedProtocolMessageType('GetEdgeFramesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEDGEFRAMESRESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetEdgeFramesResponse)
  })
_sym_db.RegisterMessage(GetEdgeFramesResponse)

GetTableFramesResponse = _reflection.GeneratedProtocolMessageType('GetTableFramesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTABLEFRAMESRESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetTableFramesResponse)
  })
_sym_db.RegisterMessage(GetTableFramesResponse)

CreateNamespaceRequest = _reflection.GeneratedProtocolMessageType('CreateNamespaceRequest', (_message.Message,), {

  'KwargsEntry' : _reflection.GeneratedProtocolMessageType('KwargsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CREATENAMESPACEREQUEST_KWARGSENTRY,
    '__module__' : 'GraphTypesService_pb2'
    # @@protoc_insertion_point(class_scope:xgt.CreateNamespaceRequest.KwargsEntry)
    })
  ,
  'DESCRIPTOR' : _CREATENAMESPACEREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.CreateNamespaceRequest)
  })
_sym_db.RegisterMessage(CreateNamespaceRequest)
_sym_db.RegisterMessage(CreateNamespaceRequest.KwargsEntry)

CreateNamespaceResponse = _reflection.GeneratedProtocolMessageType('CreateNamespaceResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATENAMESPACERESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.CreateNamespaceResponse)
  })
_sym_db.RegisterMessage(CreateNamespaceResponse)

DeleteNamespaceRequest = _reflection.GeneratedProtocolMessageType('DeleteNamespaceRequest', (_message.Message,), {

  'KwargsEntry' : _reflection.GeneratedProtocolMessageType('KwargsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DELETENAMESPACEREQUEST_KWARGSENTRY,
    '__module__' : 'GraphTypesService_pb2'
    # @@protoc_insertion_point(class_scope:xgt.DeleteNamespaceRequest.KwargsEntry)
    })
  ,
  'DESCRIPTOR' : _DELETENAMESPACEREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.DeleteNamespaceRequest)
  })
_sym_db.RegisterMessage(DeleteNamespaceRequest)
_sym_db.RegisterMessage(DeleteNamespaceRequest.KwargsEntry)

DeleteNamespaceResponse = _reflection.GeneratedProtocolMessageType('DeleteNamespaceResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETENAMESPACERESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.DeleteNamespaceResponse)
  })
_sym_db.RegisterMessage(DeleteNamespaceResponse)

GetNamespacesRequest = _reflection.GeneratedProtocolMessageType('GetNamespacesRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETNAMESPACESREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetNamespacesRequest)
  })
_sym_db.RegisterMessage(GetNamespacesRequest)

GetNamespacesResponse = _reflection.GeneratedProtocolMessageType('GetNamespacesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETNAMESPACESRESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetNamespacesResponse)
  })
_sym_db.RegisterMessage(GetNamespacesResponse)

CreateFrameRequest = _reflection.GeneratedProtocolMessageType('CreateFrameRequest', (_message.Message,), {

  'KwargsEntry' : _reflection.GeneratedProtocolMessageType('KwargsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CREATEFRAMEREQUEST_KWARGSENTRY,
    '__module__' : 'GraphTypesService_pb2'
    # @@protoc_insertion_point(class_scope:xgt.CreateFrameRequest.KwargsEntry)
    })
  ,
  'DESCRIPTOR' : _CREATEFRAMEREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.CreateFrameRequest)
  })
_sym_db.RegisterMessage(CreateFrameRequest)
_sym_db.RegisterMessage(CreateFrameRequest.KwargsEntry)

GetFramesMetadataRequest = _reflection.GeneratedProtocolMessageType('GetFramesMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFRAMESMETADATAREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetFramesMetadataRequest)
  })
_sym_db.RegisterMessage(GetFramesMetadataRequest)

GetFramesMetadataResponse = _reflection.GeneratedProtocolMessageType('GetFramesMetadataResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETFRAMESMETADATARESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetFramesMetadataResponse)
  })
_sym_db.RegisterMessage(GetFramesMetadataResponse)

DeleteFrameRequest = _reflection.GeneratedProtocolMessageType('DeleteFrameRequest', (_message.Message,), {

  'KwargsEntry' : _reflection.GeneratedProtocolMessageType('KwargsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DELETEFRAMEREQUEST_KWARGSENTRY,
    '__module__' : 'GraphTypesService_pb2'
    # @@protoc_insertion_point(class_scope:xgt.DeleteFrameRequest.KwargsEntry)
    })
  ,
  'DESCRIPTOR' : _DELETEFRAMEREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.DeleteFrameRequest)
  })
_sym_db.RegisterMessage(DeleteFrameRequest)
_sym_db.RegisterMessage(DeleteFrameRequest.KwargsEntry)

DeleteFrameResponse = _reflection.GeneratedProtocolMessageType('DeleteFrameResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETEFRAMERESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.DeleteFrameResponse)
  })
_sym_db.RegisterMessage(DeleteFrameResponse)

GetFrameSizeRequest = _reflection.GeneratedProtocolMessageType('GetFrameSizeRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFRAMESIZEREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetFrameSizeRequest)
  })
_sym_db.RegisterMessage(GetFrameSizeRequest)

GetFrameSizeResponse = _reflection.GeneratedProtocolMessageType('GetFrameSizeResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETFRAMESIZERESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetFrameSizeResponse)
  })
_sym_db.RegisterMessage(GetFrameSizeResponse)

GetFrameLabelsRequest = _reflection.GeneratedProtocolMessageType('GetFrameLabelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFRAMELABELSREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetFrameLabelsRequest)
  })
_sym_db.RegisterMessage(GetFrameLabelsRequest)

GetFrameLabelsResponse = _reflection.GeneratedProtocolMessageType('GetFrameLabelsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETFRAMELABELSRESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetFrameLabelsResponse)
  })
_sym_db.RegisterMessage(GetFrameLabelsResponse)

GetRowLabelUniverseRequest = _reflection.GeneratedProtocolMessageType('GetRowLabelUniverseRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETROWLABELUNIVERSEREQUEST,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetRowLabelUniverseRequest)
  })
_sym_db.RegisterMessage(GetRowLabelUniverseRequest)

GetRowLabelUniverseResponse = _reflection.GeneratedProtocolMessageType('GetRowLabelUniverseResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETROWLABELUNIVERSERESPONSE,
  '__module__' : 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetRowLabelUniverseResponse)
  })
_sym_db.RegisterMessage(GetRowLabelUniverseResponse)

_GRAPHTYPESSERVICE = DESCRIPTOR.services_by_name['GraphTypesService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _FRAMELABELMAP_ACCESSLABELSENTRY._options = None
  _FRAMELABELMAP_ACCESSLABELSENTRY._serialized_options = b'8\001'
  _CREATENAMESPACEREQUEST_KWARGSENTRY._options = None
  _CREATENAMESPACEREQUEST_KWARGSENTRY._serialized_options = b'8\001'
  _DELETENAMESPACEREQUEST_KWARGSENTRY._options = None
  _DELETENAMESPACEREQUEST_KWARGSENTRY._serialized_options = b'8\001'
  _CREATEFRAMEREQUEST_KWARGSENTRY._options = None
  _CREATEFRAMEREQUEST_KWARGSENTRY._serialized_options = b'8\001'
  _DELETEFRAMEREQUEST_KWARGSENTRY._options = None
  _DELETEFRAMEREQUEST_KWARGSENTRY._serialized_options = b'8\001'
  _LABELSET._serialized_start=95
  _LABELSET._serialized_end=120
  _FRAMELABELMAP._serialized_start=123
  _FRAMELABELMAP._serialized_end=267
  _FRAMELABELMAP_ACCESSLABELSENTRY._serialized_start=201
  _FRAMELABELMAP_ACCESSLABELSENTRY._serialized_end=267
  _GETVERTEXFRAMESRESPONSE._serialized_start=269
  _GETVERTEXFRAMESRESPONSE._serialized_end=382
  _GETEDGEFRAMESRESPONSE._serialized_start=384
  _GETEDGEFRAMESRESPONSE._serialized_end=491
  _GETTABLEFRAMESRESPONSE._serialized_start=493
  _GETTABLEFRAMESRESPONSE._serialized_end=603
  _CREATENAMESPACEREQUEST._serialized_start=606
  _CREATENAMESPACEREQUEST._serialized_end=860
  _CREATENAMESPACEREQUEST_KWARGSENTRY._serialized_start=797
  _CREATENAMESPACEREQUEST_KWARGSENTRY._serialized_end=860
  _CREATENAMESPACERESPONSE._serialized_start=862
  _CREATENAMESPACERESPONSE._serialized_end=914
  _DELETENAMESPACEREQUEST._serialized_start=917
  _DELETENAMESPACEREQUEST._serialized_end=1103
  _DELETENAMESPACEREQUEST_KWARGSENTRY._serialized_start=797
  _DELETENAMESPACEREQUEST_KWARGSENTRY._serialized_end=860
  _DELETENAMESPACERESPONSE._serialized_start=1105
  _DELETENAMESPACERESPONSE._serialized_end=1184
  _GETNAMESPACESREQUEST._serialized_start=1186
  _GETNAMESPACESREQUEST._serialized_end=1208
  _GETNAMESPACESRESPONSE._serialized_start=1210
  _GETNAMESPACESRESPONSE._serialized_end=1284
  _CREATEFRAMEREQUEST._serialized_start=1287
  _CREATEFRAMEREQUEST._serialized_end=1682
  _CREATEFRAMEREQUEST_KWARGSENTRY._serialized_start=797
  _CREATEFRAMEREQUEST_KWARGSENTRY._serialized_end=860
  _GETFRAMESMETADATAREQUEST._serialized_start=1684
  _GETFRAMESMETADATAREQUEST._serialized_end=1748
  _GETFRAMESMETADATARESPONSE._serialized_start=1750
  _GETFRAMESMETADATARESPONSE._serialized_end=1838
  _DELETEFRAMEREQUEST._serialized_start=1841
  _DELETEFRAMEREQUEST._serialized_end=1999
  _DELETEFRAMEREQUEST_KWARGSENTRY._serialized_start=797
  _DELETEFRAMEREQUEST_KWARGSENTRY._serialized_end=860
  _DELETEFRAMERESPONSE._serialized_start=2001
  _DELETEFRAMERESPONSE._serialized_end=2076
  _GETFRAMESIZEREQUEST._serialized_start=2078
  _GETFRAMESIZEREQUEST._serialized_end=2119
  _GETFRAMESIZERESPONSE._serialized_start=2121
  _GETFRAMESIZERESPONSE._serialized_end=2184
  _GETFRAMELABELSREQUEST._serialized_start=2186
  _GETFRAMELABELSREQUEST._serialized_end=2229
  _GETFRAMELABELSRESPONSE._serialized_start=2231
  _GETFRAMELABELSRESPONSE._serialized_end=2327
  _GETROWLABELUNIVERSEREQUEST._serialized_start=2329
  _GETROWLABELUNIVERSEREQUEST._serialized_end=2377
  _GETROWLABELUNIVERSERESPONSE._serialized_start=2379
  _GETROWLABELUNIVERSERESPONSE._serialized_end=2470
  _GRAPHTYPESSERVICE._serialized_start=2473
  _GRAPHTYPESSERVICE._serialized_end=3357
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import GraphTypesService_pb2 as GraphTypesService__pb2


class GraphTypesServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetVertexFrames = channel.unary_unary(
                '/xgt.GraphTypesService/GetVertexFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                )
        self.GetEdgeFrames = channel.unary_unary(
                '/xgt.GraphTypesService/GetEdgeFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                )
        self.GetTableFrames = channel.unary_unary(
                '/xgt.GraphTypesService/GetTableFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                )
        self.CreateNamespace = channel.unary_unary(
                '/xgt.GraphTypesService/CreateNamespace',
                request_serializer=GraphTypesService__pb2.CreateNamespaceRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.CreateNamespaceResponse.FromString,
                )
        self.DeleteNamespace = channel.unary_unary(
                '/xgt.GraphTypesService/DeleteNamespace',
                request_serializer=GraphTypesService__pb2.DeleteNamespaceRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.DeleteNamespaceResponse.FromString,
                )
        self.GetNamespaces = channel.unary_unary(
                '/xgt.GraphTypesService/GetNamespaces',
                request_serializer=GraphTypesService__pb2.GetNamespacesRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetNamespacesResponse.FromString,
                )
        self.CreateFrame = channel.unary_unary(
                '/xgt.GraphTypesService/CreateFrame',
                request_serializer=GraphTypesService__pb2.CreateFrameRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                )
        self.DeleteFrame = channel.unary_unary(
                '/xgt.GraphTypesService/DeleteFrame',
                request_serializer=GraphTypesService__pb2.DeleteFrameRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.DeleteFrameResponse.FromString,
                )
        self.GetFrameSize = channel.unary_unary(
                '/xgt.GraphTypesService/GetFrameSize',
                request_serializer=GraphTypesService__pb2.GetFrameSizeRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFrameSizeResponse.FromString,
                )
        self.GetFrameLabels = channel.unary_unary(
                '/xgt.GraphTypesService/GetFrameLabels',
                request_serializer=GraphTypesService__pb2.GetFrameLabelsRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFrameLabelsResponse.FromString,
                )
        self.GetRowLabelUniverse = channel.unary_unary(
                '/xgt.GraphTypesService/GetRowLabelUniverse',
                request_serializer=GraphTypesService__pb2.GetRowLabelUniverseRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetRowLabelUniverseResponse.FromString,
                )


class GraphTypesServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetVertexFrames(self, request, context):
        """Vertex frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEdgeFrames(self, request, context):
        """Edge frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTableFrames(self, request, context):
        """Table frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateNamespace(self, request, context):
        """Namespaces functions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteNamespace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNamespaces(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateFrame(self, request, context):
        """Multi-type functions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteFrame(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFrameSize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFrameLabels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRowLabelUniverse(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GraphTypesServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetVertexFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVertexFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'GetEdgeFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEdgeFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'GetTableFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTableFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'CreateNamespace': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateNamespace,
                    request_deserializer=GraphTypesService__pb2.CreateNamespaceRequest.FromString,
                    response_serializer=GraphTypesService__pb2.CreateNamespaceResponse.SerializeToString,
            ),
            'DeleteNamespace': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteNamespace,
                    request_deserializer=GraphTypesService__pb2.DeleteNamespaceRequest.FromString,
                    response_serializer=GraphTypesService__pb2.DeleteNamespaceResponse.SerializeToString,
            ),
            'GetNamespaces': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNamespaces,
                    request_deserializer=GraphTypesService__pb2.GetNamespacesRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetNamespacesResponse.SerializeToString,
            ),
            'CreateFrame': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateFrame,
                    request_deserializer=GraphTypesService__pb2.CreateFrameRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'DeleteFrame': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteFrame,
                    request_deserializer=GraphTypesService__pb2.DeleteFrameRequest.FromString,
                    response_serializer=GraphTypesService__pb2.DeleteFrameResponse.SerializeToString,
            ),
            'GetFrameSize': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFrameSize,
                    request_deserializer=GraphTypesService__pb2.GetFrameSizeRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFrameSizeResponse.SerializeToString,
            ),
            'GetFrameLabels': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFrameLabels,
                    request_deserializer=GraphTypesService__pb2.GetFrameLabelsRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFrameLabelsResponse.SerializeToString,
            ),
            'GetRowLabelUniverse': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRowLabelUniverse,
                    request_deserializer=GraphTypesService__pb2.GetRowLabelUniverseRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetRowLabelUniverseResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'xgt.GraphTypesService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class GraphTypesService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetVertexFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetVertexFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEdgeFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetEdgeFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTableFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetTableFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateNamespace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/CreateNamespace',
            GraphTypesService__pb2.CreateNamespaceRequest.SerializeToString,
            GraphTypesService__pb2.CreateNamespaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteNamespace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/DeleteNamespace',
            GraphTypesService__pb2.DeleteNamespaceRequest.SerializeToString,
            GraphTypesService__pb2.DeleteNamespaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetNamespaces(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetNamespaces',
            GraphTypesService__pb2.GetNamespacesRequest.SerializeToString,
            GraphTypesService__pb2.GetNamespacesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateFrame(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/CreateFrame',
            GraphTypesService__pb2.CreateFrameRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteFrame(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/DeleteFrame',
            GraphTypesService__pb2.DeleteFrameRequest.SerializeToString,
            GraphTypesService__pb2.DeleteFrameResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFrameSize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetFrameSize',
            GraphTypesService__pb2.GetFrameSizeRequest.SerializeToString,
            GraphTypesService__pb2.GetFrameSizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFrameLabels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetFrameLabels',
            GraphTypesService__pb2.GetFrameLabelsRequest.SerializeToString,
            GraphTypesService__pb2.GetFrameLabelsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRowLabelUniverse(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/xgt.GraphTypesService/GetRowLabelUniverse',
            GraphTypesService__pb2.GetRowLabelUniverseRequest.SerializeToString,
            GraphTypesService__pb2.GetRowLabelUniverseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2022 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import logging

import grpc

import pyarrow

from . import DataService_pb2 as data_proto
from . import ErrorMessages_pb2 as err_proto

log = logging.getLogger(__name__)

BOOLEAN = 'boolean'
INT = 'int'
FLOAT = 'float'
DATE = 'date'
TIME = 'time'
DATETIME = 'datetime'
IPADDRESS = 'ipaddress'
TEXT = 'text'
LIST = 'list'

# Send in 2MB chunks (grpc recommends 16-64 KB, but this got the best performance locally)
# FYI: by default grpc only supports up to 4MB.
MAX_PACKET_SIZE = 2097152

class XgtError(Exception):
  """
  Base exception class from which all other xgt exceptions inherit. It is
  raised in error cases that don't have a specific xgt exception type.
  """
  def __init__(self, msg, trace=''):
    self.msg = msg
    self.trace = trace

    if log.getEffectiveLevel() >= logging.DEBUG:
      if self.trace != '':
        log.debug(self.trace)
      else:
        log.debug(self.msg)
    Exception.__init__(self, self.msg)

class XgtNotImplemented(XgtError):
  """Raised for functionality with pending implementation."""
class XgtInternalError(XgtError):
  """
  Intended for internal server purposes only. This exception should not become
  visible to the user.
  """
class XgtIOError(XgtError):
  """An I/O problem occurred either on the client or server side."""
  def __init__(self, msg, trace='', job = None):
    self._job = job
    XgtError.__init__(self, msg, trace)

  @property
  def job(self):
    """Job: Job associated with the load/insert operation if available. May be None."""
    return self._job
class XgtServerMemoryError(XgtError):
  """
  The server memory usage is close to or at capacity and work could be lost.
  """
class XgtConnectionError(XgtError):
  """
  The client cannot properly connect to the server. This can include a failure
  to connect due to an xgt module version error.
  """
class XgtSyntaxError(XgtError):
  """A query was provided with incorrect syntax."""
class XgtTypeError(XgtError):
  """
  An unexpected type was supplied.

  For queries, an invalid data type was used either as an entity or as a
  property. For frames, either an edge, vertex or table frames was expected
  but the wrong frame type or some other data type was provided. For
  properties, the property declaration establishes the expected data type. A
  type error is raise if the data type used is not appropriate.
  """
class XgtValueError(XgtError):
  """An invalid or unexpected value was provided."""
class XgtNameError(XgtError):
  """
  An unexpected name was provided. Typically can occur during object retrieval
  where the object name was not found.
  """
class XgtArithmeticError(XgtError):
  """An invalid arithmetic calculation was detected and cannot be handled."""
class XgtFrameDependencyError(XgtError):
  """
  The requested action will produce an invalid graph or break a valid graph.
  """
class XgtTransactionError(XgtError):
  """A Transaction was attempted but didn't complete."""
class XgtSecurityError(XgtError):
  """A security violation occured."""

# Validation support functions

def _validated_schema(obj):
  '''Takes a user-supplied object and returns a valid schema.

  Users can supply a variety of objects as valid schemas. To simplify internal
  processing, we canonicalize these into a list of string-type pairs,
  performing validation along the way.
  '''
  # Validate the shape first
  try:
    if len(obj) < 1:
      raise XgtTypeError('A schema must not be empty.')
    for col in obj:
      assert len(col) >= 2
      if (_validated_property_type(col[1]) == "LIST"):
        assert (len(col) <= 4 and len(col) >= 3)
      else:
        assert len(col) == 2
  except:
    raise XgtTypeError('A schema must be a non-empty list of (property, type) pairs.')
  # Looks good. Return a canonical schema.
  schema_returned = []
  for col in obj:
    val_type = _validated_property_type(col[1])
    if val_type != "LIST":
      schema_returned.append((_validated_property_name(col[0]), val_type))
    else:
      leaf_type = _validated_property_type(col[2])
      if (len(col) != 4):
        schema_returned.append((_validated_property_name(col[0]),
                               val_type, leaf_type))
      else:
        schema_returned.append((_validated_property_name(col[0]),
                                val_type, leaf_type, col[3]))
  return schema_returned

def _validated_frame_name(obj):
  '''Takes a user-supplied object and returns a unicode frame name string.'''
  _assert_isstring(obj)
  name = str(obj)
  if len(name) < 1:
    raise XgtNameError('Frame names cannot be empty.')
  return name

def _validated_namespace_name(obj):
  '''Takes a user-supplied object and returns a unicode frame name string.'''
  _assert_isstring(obj)
  name = str(obj)
  if len(name) < 1:
    raise XgtNameError('Namespace names cannot be empty.')
  return name

def _validated_property_name(obj):
  '''Takes a user-supplied object and returns a unicode property name string.'''
  _assert_isstring(obj)
  return str(obj)

def _get_valid_property_types_to_create():
  return [BOOLEAN, INT, FLOAT, DATE, TIME, DATETIME, IPADDRESS, TEXT, LIST]

def _get_valid_property_types_for_return_only():
  return ['container_id', 'job_id']

def _validated_property_type(obj):
  '''Takes a user-supplied object and returns an xGT schema type.'''
  _assert_isstring(obj)
  prop_type = str(obj)
  valid_prop_types = _get_valid_property_types_to_create()
  if prop_type.lower() not in valid_prop_types:
    if prop_type.lower in _get_valid_property_types_for_return_only():
      raise XgtTypeError('Invalid property type "'+prop_type+'". This type '
                         'cannot be used when creating a frame.')
    else:
      raise XgtTypeError('Invalid property type "'+prop_type+'"')
  return prop_type.upper()

def _validate_opt_level(optlevel):
  """
  Valid optimization level values are:
    - 0: No optimization.
    - 1: General optimization.
    - 2: WHERE-clause optimization.
    - 3: Degree-cycle optimization.
    - 4: Query order optimization.
  """
  if isinstance(optlevel, int):
    if optlevel not in [0, 1, 2, 3, 4]:
      raise XgtValueError("Invalid optlevel '" + str(optlevel) +"'")
  else:
    raise XgtTypeError("optlevel must be an integer")
  return True

def _assert_noerrors(response):
  if len(response.error) > 0:
    error = response.error[0]
    try:
      error_code_name = err_proto.ErrorCodeEnum.Name(error.code)
      error_class = _code_error_map[error_code_name]
      raise error_class(error.message, error.detail)
    except XgtError:
      raise
    except Exception as ex:
      raise XgtError("Error detected while raising exception" +
                     str(ex), str(ex))

def _convert_flight_server_error_into_xgt(error):
  if len(error.extra_info) >= 8 and error.extra_info[0:6] == b"ERROR:":
    try:
      error_class = _code_error_map[err_proto.ErrorCodeEnum.Name(int(error.extra_info[6:8]))]
      return error_class(str(error), error.extra_info)
    except:
      pass
  return XgtError(str(error))

def _assert_isstring(value):
  if not isinstance(value, str):
    msg = str(value) + " is not a string"
    raise TypeError(msg)

_code_error_map = {
  'GENERIC_ERROR': XgtError,
  'NOT_IMPLEMENTED': XgtNotImplemented,
  'INTERNAL_ERROR': XgtInternalError,
  'IO_ERROR': XgtIOError,
  'SERVER_MEMORY_ERROR': XgtServerMemoryError,
  'CONNECTION_ERROR': XgtConnectionError,
  'SYNTAX_ERROR': XgtSyntaxError,
  'TYPE_ERROR': XgtTypeError,
  'VALUE_ERROR': XgtValueError,
  'NAME_ERROR': XgtNameError,
  'ARITHMETIC_ERROR': XgtArithmeticError,
  'FRAME_DEPENDENCY_ERROR': XgtFrameDependencyError,
  'TRANSACTION_ERROR': XgtTransactionError,
  'SECURITY_ERROR': XgtSecurityError,
}

def _create_flight_ticket(name, offset, length, include_row_labels,
                          row_label_column_header = None,
                          order = True, date_as_string = True,
                          job_id = None):
  if isinstance(offset, str):
    offset = int(offset)
  elif not isinstance(offset, int):
    raise ValueError("offset must be an int.")
  if offset < 0:
    raise ValueError("offset must be >= 0.")

  if length is not None:
    if isinstance(length, str):
      length = int(length)
    elif not isinstance(length, int):
      raise ValueError("length must be an int.")
    if length < 0:
      raise ValueError("length must be >= 0.")

  ticket = '`' + name + '`'

  if offset != 0:
    ticket += ".offset=" + str(offset)
  if length != None:
    ticket += ".length=" + str(length)
  if order:
    ticket += ".order=True"
  if date_as_string:
    ticket += ".dates_as_strings=True"
  if include_row_labels:
    ticket += ".egest_row_labels=True"
  if row_label_column_header is not None:
    ticket += ".label_column_header=" + row_label_column_header

  if job_id is not None:
    if isinstance(job_id, str):
      job_id = int(job_id)
    elif not isinstance(job_id, int):
      raise ValueError("job ID must be an int.")
    ticket += ".job_id=" + str(job_id)

  return ticket

def _get_data_python_from_table(arrow_table):
  # List comprehension here is simpler, but has slow performance due to the
  # access pattern being bad for the cache hits.
  return_list = [None] * arrow_table.num_rows
  for i in range(arrow_table.num_rows):
    return_list[i] = []
  for i, x in enumerate(arrow_table):
    for j, y in enumerate(x):
      return_list[j].append(y.as_py())
  return return_list

def _get_data_arrow(conn, ticket):
  try:
    res_table = conn.arrow_conn.do_get(pyarrow.flight.Ticket(ticket)).read_all()
    return res_table
  except pyarrow._flight.FlightServerError as err:
    raise _convert_flight_server_error_into_xgt(err) from err
  except pyarrow._flight.FlightUnavailableError as err:
    raise XgtConnectionError(str(err)) from err

def _get_data_python(conn, ticket):
  res_table = _get_data_arrow(conn, ticket)
  return _get_data_python_from_table(res_table)

def _get_data_pandas(conn, ticket):
  res_table = _get_data_arrow(conn, ticket)
  return res_table.to_pandas()

# Helper functions for low code.

# Convert the pyarrow type to an xgt type.
def _pyarrow_type_to_xgt_type(pyarrow_type):
  if pyarrow.types.is_boolean(pyarrow_type):
    return BOOLEAN
  elif pyarrow.types.is_timestamp(pyarrow_type) or pyarrow.types.is_date64(pyarrow_type):
    return DATETIME
  elif pyarrow.types.is_date(pyarrow_type):
    return DATE
  elif pyarrow.types.is_time(pyarrow_type):
    return TIME
  elif pyarrow.types.is_integer(pyarrow_type):
    return INT
  elif pyarrow.types.is_float32(pyarrow_type) or \
       pyarrow.types.is_float64(pyarrow_type) or \
       pyarrow.types.is_decimal(pyarrow_type):
    return FLOAT
  elif pyarrow.types.is_string(pyarrow_type):
    return TEXT
  else:
    raise XgtTypeError("Cannot convert pyarrow type " + str(pyarrow_type) + " to xGT type.")

def _infer_xgt_schema_from_pyarrow_schema(pyarrow_schema):
  return [[c.name, _pyarrow_type_to_xgt_type(c.type)] for c in pyarrow_schema]

# Get the column in the schema by name or position.
def _find_key_in_schema(key, schema):
  if isinstance(key, str):
    found_key = False
    for elem in schema:
      if elem[0] == key:
        found_key = True
        break
    if not found_key:
      raise XgtNameError("The key " + str(key) + " not found in schema.")
    return key
  elif isinstance(key, int):
    if key >= len(schema) or key < 0:
      raise XgtError("Could not locate key " + str(key) + " in schema with " + str(len(schema)) + " entries.")
    return schema[key][0]

# Modify an xgt schema based on a frame column name to data column name
# mapping. The key names of this map will become the column names of the
# new schema. The values of the map correspond to the columns of the
# initial schema.
def _apply_mapping_to_schema(initial_schema, frame_to_data_column_mapping):
  def find_data_col_name(data_col):
    if isinstance(data_col, str):
      return data_col
    elif isinstance(data_col, int):
      if data_col >= len(initial_schema) or data_col < 0:
        err("Error creating the schema. The column mapping refers to data column "
            "position " + str(data_col) + ", but only " + str(len(initial_schema)) +
            " columns were found in the data.")
        raise XgtValueError(err)
      return initial_schema[data_col][0]

  data_col_name_to_type = { elem[0] : elem[1] for elem in initial_schema }

  schema = []
  for frame_col, data_col in frame_to_data_column_mapping.items():
    data_type = data_col_name_to_type[find_data_col_name(data_col)]
    schema.append([frame_col, data_type])

  return schema

def _remove_label_columns_from_schema(initial_schema, row_label_columns):
  def find_data_col_name(data_col):
    if isinstance(data_col, str):
      return data_col
    elif isinstance(data_col, int):
      if data_col >= len(initial_schema) or data_col < 0:
        err("Error creating the schema. The row_label_columns parameter refers to data column "
            "position " + str(data_col) + ", but only " + str(len(initial_schema)) +
            " columns were found in the data.")
        raise XgtValueError(err)
      return initial_schema[data_col][0]

  data_col_name_to_type = { elem[0] : elem[1] for elem in initial_schema }

  label_columns = set([find_data_col_name(col) for col in row_label_columns])

  return [col for col in initial_schema if col[0] not in label_columns]



# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2022 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

from datetime import datetime
import logging
import json

from . import ErrorMessages_pb2 as err_proto
from . import JobService_pb2 as job_proto
from . import SchemaMessages_pb2 as sch_proto
from .common import (XgtError, XgtInternalError, XgtNotImplemented,
                     _create_flight_ticket, _get_data_python_from_table,
                     _get_data_arrow, _code_error_map)

log = logging.getLogger(__name__)

class _QueryPlan(object):
  class _QueryPlanSingleFrame(object):
    def __init__(self, query_element_message):
      self._label = query_element_message.source_label
      self._frame = query_element_message.source_frame

    @property
    def name(self):
      return self._label + ":" + self._frame;

    def __str__(self):
      return "(" + self._label + ":" + self._frame + ")"

  class _QueryPlanEdge(object):
    def __init__(self, query_edge_message):
      if query_edge_message.edge_direction == \
         sch_proto.QueryEdgeDirection.Value('FORWARD'):
        self.forward_edge = True
      else:
        self.forward_edge = False
      self._source_label = query_edge_message.source_label
      self._source_frame = query_edge_message.source_frame
      self._edge_label = query_edge_message.edge_label
      self._edge_frame = query_edge_message.edge_frame
      self._target_label = query_edge_message.target_label
      self._target_frame = query_edge_message.target_frame

    @property
    def name(self):
      return self._edge_label + ":" + self._edge_frame;

    @property
    def edge_name(self):
      return self._edge_label + ":" + self._edge_frame;

    @property
    def source_name(self):
      return self._source_label + ":" + self._source_frame;

    @property
    def target_name(self):
      return self._target_label + ":" + self._target_frame;

    # The string representation of a _QueryPlanEdge object is a tuple
    # of the source vertex, edge, and target vertex. The direction
    # of the edge is always forward.
    def __str__(self):
      if self.forward_edge:
        vertex1 = self.source_name
        vertex2 = self.target_name
      else:
        vertex2 = self.source_name
        vertex1 = self.target_name
      return "(" + vertex1 + ", " + self.edge_name + ", " + vertex2 + ")"

  def __init__(self, query_plan_response):
    # The _plan member is a list of lists. Each list represents a query
    # that starts with MATCH in cypher and contains elements representing
    # edges or (if only a single frame was matched), a vertex or table.
    self._plan = []
    for query in query_plan_response.plan_as_edge_list:
      query_representation = []
      plan_score = None
      for element in query.edges:
        if element.edge_direction == \
           sch_proto.QueryEdgeDirection.Value('NOT_EDGE'):
          query_representation.append(self._QueryPlanSingleFrame(element))
        else:
          query_representation.append(self._QueryPlanEdge(element))
      if (query.HasField("plan_score")):
        plan_score = query.plan_score
      self._plan.append((query_representation, plan_score))

  def __str__(self):
    as_string = ""
    for (idx, query) in enumerate(self._plan):
      if idx > 0:
        as_string += "\n"
      as_string += "QUERY:"
      for element in query[0]:
        as_string += "\n" + str(element)
    return as_string

  def to_cypher(self):
    cypher = ""
    for query, score in self._plan:
      # The corresponding query section had no MATCH portion. It won't be
      # included in the cypher representation currently used as only
      # the structure of the MATCH portions is shown.
      if len(query) == 0:
        continue
      if len(cypher) > 0:
        cypher += "\n"
      cypher += "MATCH\n"
      last_target_name = ""

      # Write the first element of the query. This must be a vertex or table
      # frame.
      if isinstance(query[0], self._QueryPlanEdge):
        cypher += "(" + query[0].source_name + ")"
      else:
        cypher += "(" + query[0].name + ")"

      for (element_idx, element) in enumerate(query):
        if isinstance(element, self._QueryPlanEdge):
          if last_target_name != element.source_name and element_idx > 0:
            # This edge starts a new path in the query. We must write the
            # source.
            cypher += ",\n(" + element.source_name + ")"
          if not element.forward_edge:
            cypher += "<"
          cypher += "-[" + element.edge_name + "]-"
          if element.forward_edge:
            cypher += ">"
          cypher += "(" + element.target_name + ")"
          last_target_name = element.target_name
        elif element_idx > 0:
          # This is not an edge element. It is a vertex or table frame.
          # If it was the first in this query then it has already been
          # written. Otherwise, it starts a new path in the query and
          # must be written.
          cypher += ",\n(" + element.name + ")"
    return cypher

  def get_scores(self):
    return [score for query, score in self._plan]

class Job(object):
  """
  Represents a user-scheduled Job.

  An instance of this object is created by job-scheduling functions like
  `xgt.Connection.run_job` and `xgt.Connection.schedule_job`.

  A `Job` is used as a proxy for a job in the server and allows the user
  to monitor its execution, possibly cancel it, and learn about its status
  during and after execution.

  The conn parameter represents a previously established connection to the xGT
  server.

  The job_response parameter is a single element of the array returned by the
  output of a job creation gRPC call.  Each individual element in the array will
  be constructed as a separate `Job` object.

  """
  def __init__(self, conn, job_response = None, python_results = None,
               python_errors = None):
    """
    Constructor for Job. Called when Job is created.
    """
    self._conn = conn
    if job_response is not None:
      self._id = job_response.job_id
      self._user = job_response.user
      self._data = self._parse_job_data(job_response)
    else:
      self._id = 0
      self._user = ""
      self._data = { 'status' : 'completed', 'total_rows' : 0 }
    if python_results is not None:
      self._data['python_rows'] = python_results
      self._data['total_rows'] += len(python_results)
    if python_errors is not None:
      if 'ingest_errors' in self._data.keys():
        self._data['ingest_errors'] += python_errors
        self._data['total_ingest_errors'] += len(python_errors)
      else:
        if len(python_errors) > 0:
          self._data['ingest_errors'] = python_errors
        self._data['total_ingest_errors'] = len(python_errors)
    self._arrow_table = None

  def _parse_job_data(self, job_response):
    job_data = {
      'jobid': job_response.job_id,
      'user': job_response.user,
      'status': sch_proto.JobStatusEnum.Name(job_response.status).lower(),
      'start_time': job_response.start_time.ToDatetime().isoformat(),
      'end_time': job_response.end_time.ToDatetime().isoformat(),
      'error_type': None,
      'visited_edges': job_response.visited_edges,
      'timing': job_response.timing,
      'description': job_response.description,
      'total_rows' : job_response.total_rows,
      'total_ingest_errors' : job_response.total_ingest_errors,
    }
    # Optional fields.
    if (job_response.HasField('schema')):
      job_data['schema'] = [[prop.name,
                             sch_proto.UvalTypeEnum.Name(prop.data_type).lower()]
                            for prop in job_response.schema.property]
    if job_response.ingest_error and len(job_response.ingest_error):
      job_data['ingest_errors'] = [data for data in job_response.ingest_error]
    if job_response.error and len(job_response.error) > 0:
        error_code_name = err_proto.ErrorCodeEnum.Name(job_response.error[0].code)
        job_data['error_type'] = _code_error_map[error_code_name]
        job_data['error'] = ', '.join([e.message for e in job_response.error])
        job_data['trace'] = ', '.join([e.detail for e in job_response.error])
    if job_response.HasField('query_plan'):
      job_data['query_plan'] =  _QueryPlan(job_response.query_plan)

    return job_data

  def _get_job_data(self):
    request = job_proto.GetJobsRequest()
    request.job_id.extend([self._id])
    responses = self._conn._call(request, self._conn._job_svc.GetJobs)
    job_data = None
    resp_cnt = 0
    for response in responses: # Expect only one response.
      job_data = self._parse_job_data(response.job_status) # Retrieve the job status.
      resp_cnt += 1

    if (resp_cnt > 1):
      raise XgtInternalError("Expected a single job in response")

    # If the status is unknown, only update the status, not other fields as
    # they will be invalid.
    returned_status = job_data['status']
    if returned_status == 'unknown_job_status':
      self._data['status'] = returned_status
      return job_data

    # Cache the response from the server.
    self._data = job_data

    if log.getEffectiveLevel() >= logging.DEBUG:
      job_id = job_data['jobid']
      job_status = job_data['status']
      user = job_data['user']
      if 'error' in job_data:
        error = job_data['error']
      else:
        error = ''
      if 'trace' in job_data:
        trace = job_data['trace']
      else:
        trace = ''
      msg = 'Job: {0} User: {1} Status: {2}'.format(str(job_id),
                                                    user,
                                                    job_status)
      if error != '':
        msg = msg + "\nError: \n" + error
      if trace != '':
        msg = msg + "\nTrace: \n" + trace
      log.debug(msg)

    return job_data

  def _is_status_final(self):
    if 'status' in self._data:
      curr_status = self._data['status']

      if ((curr_status == 'completed') or (curr_status == 'canceled') or
          (curr_status == 'failed') or (curr_status == 'rollback') or
          (curr_status == 'unknown_job_status')):
        return True

    return False

  def _get_data_helper(self, offset, length, as_pandas):
    if isinstance(offset, str):
      offset = int(offset)
    if isinstance(length, str):
      length = int(length)
    if isinstance(offset, int):
      if offset < 0:
        raise ValueError('offset is negative')
    if isinstance(length, int):
      if length < 0:
        raise ValueError('length is negative')
    if 'rows' in self._data and not as_pandas:
      total_length = len(self._data['rows'])
      if length is None:
        length = total_length
      elif not isinstance(length, int):
        raise ValueError('length must be a non-negative integer')

      rows = self._data['rows']
      startpos = min(offset, total_length)
      endpos = min(offset + length, total_length)
      return rows[startpos:endpos]
    elif 'python_rows' in self._data and not as_pandas:
      return self._data['python_rows']
    if 'pandas' in self._data and as_pandas:
      if length is None:
        total_length = len(self._data['pandas'].index)
        length = total_length
      startpos = min(offset, total_length)
      endpos = min(offset + length + 1, total_length)
      return self._data['pandas'][startpos:endpos]
    else:
      return None

  def _load_data_helper(self, as_pandas):
    if self._is_status_final():
      curr_status = self._data['status']

      if curr_status != 'completed':
        return

      # Skip getting the Arrow table if there are no results.
      # The schema was not set for jobs without immediate results.
      if 'schema' not in self._data:
        return

      if self._arrow_table is None:
        ticket = _create_flight_ticket('xgt__Job_History', offset = 0,
                                       length = None, include_row_labels = False,
                                       job_id = self._id)
        # Cache the results in an Arrow table to be able to return them to
        # the user in both Python and Pandas formats.
        self._arrow_table = _get_data_arrow(self._conn, ticket)
      if not as_pandas:
        self._data['rows'] = _get_data_python_from_table(self._arrow_table)
      else:
        self._data['pandas'] = self._arrow_table.to_pandas()

  @property
  def id(self):
    """
    int: Identifier of the job.

    A 64-bit integer value that uniquely identifies a job. It is
    automatically incremented for each scheduled job over the lifetime of
    the xGT server process.

    """
    return self._id

  @property
  def user(self):
    """
    str: User that ran the job.

    A string that identifies the user who ran the job.
    """
    return self._user

  @property
  def status(self):
    """
    str: Status of the job.

  ==================  ===============================================
        Job status
  -------------------------------------------------------------------
           Status                       Description
  ==================  ===============================================
           scheduled  The state after the job has been created, but
                      before it has started running.
             running  The job is being executed.
           completed  The job finished successfully.
            canceled  The job was canceled.
              failed  The job failed. When the job fails the `error`
                      and `trace` properties are populated.
            rollback  The job had a transactional conflict with
                      another job and was rolled back.
  unknown_job_status  The job was not found in the job history.
  ==================  ===============================================

    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'status' in self._data:
      return self._data['status']
    else:
      return ''

  @property
  def start_time(self):
    """
    str: Date and time when the job was scheduled.

    This is a formatted string that has a resolution of seconds.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'start_time' in self._data:
      if "." in self._data['start_time']:
        return datetime.strptime(self._data['start_time'], '%Y-%m-%dT%H:%M:%S.%f')
      return datetime.strptime(self._data['start_time'], '%Y-%m-%dT%H:%M:%S')
    else:
      return ''

  @property
  def end_time(self):
    """
    str: Date and time when the job finished running.

    This is a formatted string that has a resolution of seconds.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'end_time' in self._data:
      if "." in self._data['start_time']:
        return datetime.strptime(self._data['end_time'], '%Y-%m-%dT%H:%M:%S.%f')
      return datetime.strptime(self._data['end_time'], '%Y-%m-%dT%H:%M:%S')
    else:
      return ''

  @property
  def visited_edges(self):
    """
    dict: A dictionary mapping Cypher bound variable names to an integer giving
    the number of edges visited during the job for the Edge Frame referenced by
    the bound variable.

    An edge is "visited" when the query considers the edge as a match to one of
    the query path edges.  Multiple Cypher variables can refer to the same edge
    frame.

    Consider the query path
    ``()-[a:graph_edges1]->()-[b:graph_edges2]->()-[c:graph_edges1]->()`` with
    a visited_edges result of ``a -> 5, b -> 7, c -> 4``.  In performing the
    query 5 edges of type `a` were visited, and so on.  Notice that the total
    number of edges visited for the frame graph_edges1 is 9 while the number of
    edges visited for the frame graph_edges2 is 7.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    return self._data['visited_edges']

  @property
  def total_visited_edges(self):
    """
    int: The total number of edges traversed during the job. This is the sum
    of the counts for all edge labels returned in visited_edges.

    For the example given in the visited_edges documentation, the value of
    total_visited_edges would be 16.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    return sum(self._data['visited_edges'].values())

  @property
  def timing(self):
    """
    For internal use.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    return self._data['timing']

  @property
  def description(self):
    """
    str: A description supplied when the job was started.  Usually a query.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    return self._data['description']

  @property
  def error_type(self):
    """
    object: Class that belongs to the XgtError hierarchy that corresponds to
            the original exception type thrown that caused the Job to fail.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'error_type' in self._data:
      return self._data['error_type']
    else:
      return XgtError

  @property
  def error(self):
    """
    str: User-friendly error message describing the reason a job failed.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'error' in self._data:
      return self._data['error']
    else:
      return ''

  @property
  def trace(self):
    """
    str: Very detailed error message for a failed job.

    This error message contains the friendly error message and a stack
    strace for the code that participated in the error.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'trace' in self._data:
      return self._data['trace']
    else:
      return ''

  @property
  def schema(self):
    """
    list of lists: Gets the property names and types of the stored immediate
    results.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'schema' in self._data:
      return self._data['schema']
    else:
      return None

  @property
  def total_rows(self):
    """
    int: Gets the number of rows of the query results.  Note that this is >=
    than the number of rows of data extracted to the job.  To get the full
    results the query can be executed again using an INTO clause to a
    named results Table.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'total_rows' in self._data:
      return self._data['total_rows']
    else:
      return None

  @property
  def total_ingest_errors(self):
    """
    int: Get the number of errors that were thrown during ingest.
    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'total_ingest_errors' in self._data:
      return self._data['total_ingest_errors']
    else:
      return None

  def get_data(self, offset=0, length=None):
    """
    Returns immediate results data starting at a given offset and spanning a given
    length.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.

    Returns
    -------
    list of lists
    If the job object represents a TQL query with no RETURN clause, this returns None.

    """
    if (not self._is_status_final()):
      self._get_job_data()

    self._load_data_helper(as_pandas = False)
    return self._get_data_helper(offset, length, as_pandas = False)

  def get_ingest_errors(self, offset=0, length=None):
    """
    Returns a table of strings giving error information from ingest.  The first
    thousand errors raised are retrievable this way.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows of errors to be retieved starting from the row
      indicated by offset.  A value of 'None' means 'all rows' on and after
      the offset.
      Optional. Default-None.

    Returns
    -------
    list of strings
    If this is not an ingest job or no errors were raised, this returns None.

    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'ingest_errors' in self._data:
      if isinstance(offset, str):
        offset = int(offset)
      if isinstance(length, str):
        length = int(length)
      if isinstance(offset, str):
        if offset < 0:
          raise ValueError('offset is negative')
      if isinstance(length, int):
        if length < 0:
          raise ValueError('length is negative')
      total_length = len(self._data['ingest_errors'])
      if length is None:
        length = total_length
      elif not isinstance(length, int):
        raise ValueError('length must be a non-negative integer')

      errors = self._data['ingest_errors']
      startpos = min(offset, total_length)
      endpos = min(offset + length + 1, total_length)

      return errors[startpos:endpos]
    else:
      return None

  def get_data_pandas(self, offset=0, length=None):
    """
    Returns a Pandas DataFrame containing immediate results data starting at a
    given offset and spanning a given length.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.

    Returns
    -------
    Pandas DataFrame

    """
    import pandas
    if (not self._is_status_final()):
      self._get_job_data()

    self._load_data_helper(as_pandas = True)
    return self._get_data_helper(offset, length, as_pandas = True)

  def __str__(self):
    txt = ('id:{0}, user:{1}, status:{2}').format(self.id, self.user, self.status)
    if len(self.error) > 0:
      txt = txt + (', nerror:{0}').format(self.error)
    return txt

  @property
  def _query_plan(self):
    """
    For internal use.

    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'query_plan' in self._data:
      return str(self._data['query_plan'])
    else:
      return ''

  @property
  def _query_plan_as_cypher(self):
    """
    For internal use.

    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'query_plan' in self._data:
      return self._data['query_plan'].to_cypher()
    else:
      return ''

  @property
  def _query_plan_scores(self):
    """
    For internal use.

    """
    if (not self._is_status_final()):
      self._get_job_data()

    if 'query_plan' in self._data:
      return self._data['query_plan'].get_scores()
    else:
      return None
